"""Defines basic concept related to outcomes

Outcomes in this package are always assumed to be multi-issue outcomes where  single-issue outcomes can be implemented
as the special case with a single issue.

- Both Continuous and discrete issues are supported. All issue will have names. If none is given, a random name will be
  used. It is HIGHLY recommended to always name your issues.
- Outcomes are dictionaries with issue names as keys and issue values as values.

Examples:

  Different ways to create issues:

  >>> issues = [Issue((0.5, 2.0), 'price'), Issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')
  ...           , Issue(20, 'count')]
  >>> for _ in issues: print(_)
  price: (0.5, 2.0)
  date: ['2018.10.1', '2018.10.2', '2018.10.3']
  count: 20

  Outcome example compatible with the given set of issues:

  >>> a = {'price': 1.2, 'date': '2018.10.04', 'count': 4}

"""

import itertools
import math
import random
import xml.etree.ElementTree as ET
from enum import Enum
from functools import reduce
from operator import mul
from typing import Optional, Collection, List, Generator, Iterable, Sequence, Union, Type
from typing import Tuple, Mapping, Dict, Any

import numpy as np
import pkg_resources
from dataclasses import dataclass, fields

from .java import PYTHON_CLASS_IDENTIFIER
from .common import NamedObject
from .generics import *
from .helpers import unique_name

LARGE_NUMBER = 100
__all__ = [
    'Outcome',
    'OutcomeType',
    'OutcomeRange',
    'ResponseType',
    'Issue',
    'Issues',
    'outcome_is_valid',
    'outcome_is_complete',
    'outcome_range_is_valid',
    'outcome_range_is_complete',
    'outcome_in_range',
    'enumerate_outcomes',
    'sample_outcomes',
    'outcome_as_dict',
    'outcome_as_tuple',
    'num_outcomes',
]


class ResponseType(Enum):
    """Possible answers to offers during negotiation."""
    ACCEPT_OFFER = 0
    REJECT_OFFER = 1
    END_NEGOTIATION = 2
    NO_RESPONSE = 3


class Issue(NamedObject):
    """Encodes an Issue.

    Args:
            values: Possible values for the issue
            name: Name of the issue. If not given, a random name will be generated

    Examples:

        >>> print(Issue(['to be', 'not to be'], name='THE problem'))
        THE problem: ['to be', 'not to be']
        >>> print(Issue(3, name='Cars'))
        Cars: 3
        >>> print(Issue((0.0, 1.0), name='required accuracy'))
        required accuracy: (0.0, 1.0)
        >>> a = Issue((0.0, 1.0), name='required accuracy')
        >>> a.is_continuous()
        True
        >>> a.is_discrete()
        False

    Remarks:

        - Issues can be initialized by either an iterable of strings, an integer or a tuple of two real values with
          the following meanings:
          - ``iterable of strings``: This is an issue that can any value within the given set of values (strings)
          - ``int``: This is an issue that takes any value from 0 to the given value -1 (int)
          - ``float``: This is an issue that can take any real value between the given limits (min, max)

    """

    def __init__(
        self,
        values: Union[List[str], int, Tuple[float, float]],
        name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name)
        # if isinstance(values, int) and values <= LARGE_NUMBER:
        #    values = list(range(values))
        if isinstance(values, tuple):
            values = (float(values[0]), float(values[1]))
        self.values = values

    @classmethod
    def to_xml_str(cls, issues: List['Issue']
                   , enumerate_integer: bool = True) -> str:
        """Converts the list of issues into a well-formed xml string

        Examples:

            >>> issues = [Issue(values=10, name='i1'), Issue(values=['a', 'b', 'c'], name='i2'),
            ... Issue(values=(2.5, 3.5), name='i3')]
            >>> s = Issue.to_xml_str(issues)
            >>> print(s.strip())
            <negotiation_template>
            <utility_space number_of_issues="3">
            <objective description="" etype="objective" index="0" name="root" type="objective">
                <issue etype="discrete" index="1" name="i1" type="discrete" vtype="discrete">
                    <item index="1" value="0" cost="0" description="0">
                    </item>
                    <item index="2" value="1" cost="0" description="1">
                    </item>
                    <item index="3" value="2" cost="0" description="2">
                    </item>
                    <item index="4" value="3" cost="0" description="3">
                    </item>
                    <item index="5" value="4" cost="0" description="4">
                    </item>
                    <item index="6" value="5" cost="0" description="5">
                    </item>
                    <item index="7" value="6" cost="0" description="6">
                    </item>
                    <item index="8" value="7" cost="0" description="7">
                    </item>
                    <item index="9" value="8" cost="0" description="8">
                    </item>
                    <item index="10" value="9" cost="0" description="9">
                    </item>
                </issue>
                <issue etype="discrete" index="2" name="i2" type="discrete" vtype="discrete">
                    <item index="1" value="a" cost="0" description="a">
                    </item>
                    <item index="2" value="b" cost="0" description="b">
                    </item>
                    <item index="3" value="c" cost="0" description="c">
                    </item>
                </issue>
                <issue etype="real" index="3" name="i3" type="real" vtype="real">
                    <range lowerbound="2.5" upperbound="3.5"></range>
                </issue>
            </objective>
            </utility_space>
            </negotiation_template>

            >>> issues2, _ = Issue.from_xml_str(s)
            >>> print([_.__class__.__name__ for _ in issues2])
            ['Issue', 'Issue', 'Issue']
            >>> print(len(issues2))
            3
            >>> print([str(_) for _ in issues2])
            ["i1: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']", "i2: ['a', 'b', 'c']", 'i3: (2.5, 3.5)']
            >>> print([_.values for _ in issues2])
            [['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'], ['a', 'b', 'c'], (2.5, 3.5)]


        """
        output = f'<negotiation_template>\n<utility_space number_of_issues="{len(issues)}">\n' \
                 f'<objective description="" etype="objective" index="0" name="root" type="objective">\n'
        for indx, issue in enumerate(issues):
            if isinstance(issue.values, int):
                if enumerate_integer:
                    output += f'    <issue etype="discrete" index="{indx+1}" name="{issue.name}" type="discrete" vtype="discrete">\n'
                    for i, v in enumerate(range(issue.values)):
                        output += f'        <item index="{i+1}" value="{v}" cost="0" description="{v}">\n        </item>\n'
                    output += '    </issue>\n'
                else:
                    output += f'    <issue etype="integer" index="{indx+1}" name="{issue.name}" type="integer" vtype="integer"' \
                              f' lowerbound="0" upperbound="{issue.values-1}" />\n'
            elif isinstance(issue.values, tuple):
                output += f'    <issue etype="real" index="{indx+1}" name="{issue.name}" type="real" vtype="real">\n' \
                          f'        <range lowerbound="{issue.values[0]}" upperbound="{issue.values[1]}"></range>\n    </issue>\n'
            else:
                output += f'    <issue etype="discrete" index="{indx+1}" name="{issue.name}" type="discrete" vtype="discrete">\n'
                for i, v in enumerate(issue.values):
                    output += f'        <item index="{i+1}" value="{v}" cost="0" description="{v}">\n        </item>\n'
                output += '    </issue>\n'
        output += f'</objective>\n</utility_space>\n</negotiation_template>'
        return output

    @classmethod
    def to_genius(cls, issues: Iterable['Issue'], file_name: str, enumerate_integer: bool = True) -> None:
        """Exports a the domain issues to a GENIUS XML file.

        Args:

            issues: The issues to be exported
            file_name (str): File name to export to

        Returns:

            A List[Issue] or Dict[Issue]


        Examples:

            >>> issues, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
            >>> Issue.to_genius(issues=issues, file_name = pkg_resources.resource_filename('negmas'
            ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> issues2, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues2]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD

            - Forcing Single outcome

            >>> issues, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml')
            ...     , force_single_issue=True, keep_value_names=False, keep_issue_names=False)
            >>> print(list(issues[0].all))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]
            >>> Issue.to_genius(issues=issues, enumerate_integer=True
            ...     , file_name = pkg_resources.resource_filename('negmas'
            ...                                   , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> issues3, _ = Issue.from_genius(file_name=pkg_resources.resource_filename('negmas'
            ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
            >>> print([list(issue.all) for issue in issues3])
            [['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26']]

        Remarks:
            See ``from_xml_str`` for all the parameters

        """
        with open(file_name, 'w') as f:
            f.write(cls.to_xml_str(issues=list(issues), enumerate_integer=enumerate_integer))

    @classmethod
    def from_xml_str(cls, xml_str: str
                     , force_single_issue=False, keep_value_names=True, keep_issue_names=True, safe_parsing=True
                     , n_discretization: Optional[int] = None
                     , max_n_outcomes: int = 1e6):
        """Exports a list/dict of issues from a GENIUS XML file.

        Args:

            xml_str (str): The string containing GENIUS style XML domain issue definitions
            force_single_issue (bool): Tries to generate a MappingUtility function with a single issue which is the
            product of all issues in the input
            keep_value_names (bool): Keep names of values
            keep_issue_names (bool): Keep names of issues
            safe_parsing (bool): Turn on extra checks
            n_discretization (Optional[int]): If not None, real valued issues are discretized with the given
            number of values
            max_n_outcomes (int): Maximum number of outcomes allowed (effective only if force_single_issue is True)

        Returns:

            - List[Issue] The issues (note that issue names will be stored in the name attribute of each issue if keep_issue_names)
            - List[dict] A list of agent information dicts each contains 'agent', 'class', 'utility_file_name'

        Examples:

            >>> domain_file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml')
            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=True, keep_value_names=False, keep_issue_names=False)
            >>> issue = issues[0]
            >>> print(issue.cardinality())
            27
            >>> print(list(issue.all))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=True, keep_value_names=False, keep_issue_names=True)
            >>> print(issues[0].name)
            Laptop-Harddisk-External Monitor
            >>> print(len(issues))
            1
            >>> print(list(issues[0].all))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26]

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=True, keep_value_names=True, keep_issue_names=False)
            >>> issue = issues[0]
            >>> print(issue.cardinality())
            27
            >>> print('\\n'.join(list(issue.all)[:5]))
            Dell+60 Gb+19'' LCD
            Dell+60 Gb+20'' LCD
            Dell+60 Gb+23'' LCD
            Dell+80 Gb+19'' LCD
            Dell+80 Gb+20'' LCD

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=False, keep_value_names=True)
            >>> type(issues)
            <class 'list'>
            >>> str(issues[0])
            "0: ['Dell', 'Macintosh', 'HP']"
            >>> print([_.cardinality() for _ in issues])
            [3, 3, 3]
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD

            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=True, keep_value_names=True)
            >>> len(issues)
            3
            >>> str(issues[0])
            "Laptop: ['Dell', 'Macintosh', 'HP']"
            >>> print([_.cardinality() for _ in issues])
            [3, 3, 3]
            >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues]))
            Dell Macintosh HP
            60 Gb 80 Gb 120 Gb
            19'' LCD 20'' LCD 23'' LCD


            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=False, keep_value_names=False)
            >>> len(issues)
            3
            >>> type(issues)
            <class 'list'>
            >>> str(issues[0]).split(': ')[-1]
            '3'
            >>> print([_.cardinality() for _ in issues])
            [3, 3, 3]

            >>> domain_file_name = pkg_resources.resource_filename('negmas'
            ...                              , resource_name='tests/data/fuzzyagent/single_issue_domain.xml')
            >>> issues, _ = Issue.from_xml_str(open(domain_file_name, 'r').read()
            ... , force_single_issue=False, keep_issue_names=False, keep_value_names=False)
            >>> len(issues)
            1
            >>> type(issues)
            <class 'list'>
            >>> str(issues[0]).split(': ')[-1]
            '(10.0, 40.0)'
            >>> print([_.cardinality() for _ in issues])
            [-1]
        """
        root = ET.fromstring(xml_str)
        if safe_parsing and root.tag != 'negotiation_template':
            raise ValueError(f'Root tag is {root.tag}: negotiation_template')

        utility_space = None
        agents = []
        for child in root:
            if child.tag == 'utility_space':
                utility_space = child
                for _ in utility_space:
                    if _.tag == 'objective':
                        utility_space = _
                        break
            elif child.tag == 'agent':
                agents.append(child.attrib)

        if utility_space is None:
            if safe_parsing:
                raise ValueError(f'No objective child was found in the root')
            utility_space = root
        weights = {}
        issues = {}
        issue_info = {}
        all_discrete = True
        for child in utility_space:
            if child.tag == 'issue':
                indx = int(child.attrib['index']) - 1
                myname = child.attrib['name']
                issue_key = myname if keep_issue_names else indx
                issue_info[issue_key] = {'name': myname, 'index': indx}
                info = {'type': 'discrete', 'etype': 'discrete', 'vtype': 'discrete'}
                for a in ('type', 'etype', 'vtype'):
                    info[a] = child.attrib.get(a, info[a])
                mytype = info['type']
                if mytype == 'discrete':
                    issues[issue_key] = []
                    for item in child:
                        if item.tag == 'item':
                            item_indx = int(item.attrib['index']) - 1
                            item_name = item.attrib.get('value', None)
                            item_key = item_name if keep_value_names and item_name is not None else item_indx
                            if item_key not in issues[issue_key]:  # ignore repeated items
                                issues[issue_key].append(item_key)
                    if not keep_value_names:
                        issues[issue_key] = len(issues[issue_key])
                elif mytype in ('integer', 'real'):
                    lower, upper = child.attrib.get('lowerbound', None), child.attrib.get('upperbound', None)
                    for rng_child in child:
                        if rng_child.tag == 'range':
                            lower, upper = rng_child.attrib.get('lowerbound', lower), rng_child.attrib.get('upperbound',
                                                                                                           upper)
                    if mytype == 'integer':
                        lower, upper = int(lower), int(upper)
                        if keep_value_names:
                            issues[issue_key] = list(range(lower, upper + 1))
                        else:
                            issues[issue_key] = upper - lower + 1
                    else:
                        if n_discretization is None:
                            all_discrete = False
                            issues[issue_key] = (float(lower), float(upper))
                        else:
                            issues[issue_key] = n_discretization
                else:
                    # I should add the real-valued issues code here
                    raise ValueError(f'Unknown type: {mytype}')
            else:
                raise ValueError(f'Unknown child for objective: {child.tag}')

        for key, value in zip(ikeys(issues), ivalues(issues)):
            issues[key] = Issue(values=value,
                                name=issue_info[key]['name'] if keep_issue_names else str(issue_info[key]['index']))
        issues = list(issues.values())
        if force_single_issue:
            issue_name_ = '-'.join([_.name for _ in issues]) if keep_issue_names else '0'
            if all_discrete or n_discretization is not None:
                n_outcomes = None
                if max_n_outcomes is not None:
                    n_items = [len(list(_.alli(n=n_discretization))) for _ in issues]
                    n_outcomes = reduce(mul, n_items, 1)
                    if n_outcomes > max_n_outcomes:
                        return None, None
                if keep_value_names:
                    if len(issues) > 1:
                        all_values = itertools.product(*[[str(_) for _ in issue.alli(n=n_discretization)]
                                                         for issue_key, issue in zip(range(len(issues)), issues)])
                        all_values = list(map(lambda items: '+'.join(items), all_values))
                    else:
                        all_values = [str(_) for _ in issues[0].alli(n=n_discretization)]
                    issues = [Issue(values=all_values, name=issue_name_)]
                else:
                    if n_outcomes is None:
                        n_items = [_.cardinality() for _ in issues]
                        n_outcomes = reduce(mul, n_items, 1)
                    issues = [Issue(values=n_outcomes, name=issue_name_)]
            else:
                return None, None
        return issues, agents

    @classmethod
    def from_genius(cls, file_name: str, force_single_issue=False, keep_value_names=True, keep_issue_names=True,
                    safe_parsing=True
                    , n_discretization: Optional[int] = None
                    , max_n_outcomes: int = 1e6):
        """Imports a the domain issues from a GENIUS XML file.

        Args:

            file_name (str): File name to import from

        Returns:

            A List[Issue] or Dict[Issue]


        Examples:

            >>> issues, _ = Issue.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
            >>> print([_.name for _ in issues])
            ['Laptop', 'Harddisk', 'External Monitor']

        Remarks:
            See ``from_xml_str`` for all the parameters

        """
        with open(file_name, 'r', encoding='utf-8') as f:
            xml_str = f.read()
            return cls.from_xml_str(xml_str=xml_str, force_single_issue=force_single_issue
                                    , keep_value_names=keep_value_names, keep_issue_names=keep_issue_names
                                    , safe_parsing=safe_parsing, n_discretization=n_discretization
                                    , max_n_outcomes=max_n_outcomes)

    @staticmethod
    def n_outcomes(issues: Iterable['Issue']) -> int:
        """Returns the total number of outcomes in a set of issues. `-1` indicates infinity"""
        n = 1
        for issue in issues:
            n *= issue.cardinality()
            if n < 0:
                return -1

        return n

    @staticmethod
    def generate(
        issues: Sequence[Union[int, List[str], Tuple[float, float]]],
        counts: Optional[Sequence[int]] = None,
        names: Optional[Sequence[str]] = None,
    ) -> List['Issue']:
        """Generates a set of issues with given parameters. Each is optionally repeated

        Args:

            issues: The parameters of the issues
            counts: The number of times to repeat each of the `issues`
            names: The names to assign to the issues. If None, then string representations of integers
                   starting from zero will be used.

        Returns:
            List['Issue']: The list of issues with given conditions

        """
        one_each = counts is None
        int_names = names is None
        result = []
        nxt = 0
        for i, issue in enumerate(issues):
            count = 1 if one_each else counts[i]  # type: ignore
            for j in range(count):
                name = str(nxt) if int_names else names[i]  # type: ignore
                # if count > 1:
                #    name = name + f' {j}'
                nxt += 1
                result.append(Issue(values=issue, name=name))
        return result

    @property
    def type(self) -> str:
        """The type of the issue.

        Returns:
            str: either 'continuous' or 'discrete'

        """
        if isinstance(self.values, tuple) and isinstance(
            self.values[0], float
        ):
            return 'continuous'

        elif isinstance(self.values, int) or isinstance(self.values, list):
            return 'discrete'

        raise ValueError(
            'Unknown type. Note that a Tuple[int, int] is not allowed as a range.'
        )

    def is_continuous(self) -> bool:
        """Test whether the issue is a continuous issue

        Returns:
            bool: continuous or not

        """
        return self.type.startswith('c')

    def is_discrete(self) -> bool:
        """Test whether the issue is a discrete issue

        Returns:
            bool: discrete or not

        """
        return self.type.startswith('d')

    @property
    def all(self) -> Generator:
        """A generator that generates all possible values.

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:

            >>> from negmas.outcomes import Issue
            >>> list(Issue(5).all)
            [0, 1, 2, 3, 4]

        """
        if self.is_continuous():
            raise ValueError(
                'Cannot return all possibilities of a continuous issue'
            )

        if isinstance(self.values, int):
            yield from range(self.values)

        else:
            yield from self.values  # type: ignore

    def alli(self, n: Optional[int] = 10) -> Generator:
        """A generator that generates all possible values or samples n values for real Issues.

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:

            >>> from negmas.outcomes import Issue
            >>> list(Issue(5).all)
            [0, 1, 2, 3, 4]

        """
        if self.is_continuous():
            if n is None:
                raise ValueError('Real valued issue with no discretization value')
            yield from np.linspace(self.values[0], self.values[1], num=n, endpoint=True).tolist()

        if isinstance(self.values, int):
            yield from range(self.values)
        else:
            yield from self.values  # type: ignore

    def cardinality(self) -> int:
        """The number of possible outcomes for the issue. A negative number means infinite"""
        if isinstance(self.values, int):
            return self.values

        elif self.is_continuous():
            return -1

        return len(self.values)  # type: ignore

    def rand(self) -> Union[int, float, str]:
        """Picks a random valid value."""
        if isinstance(self.values, int):
            return random.randint(0, self.values - 1)

        elif self.is_continuous():
            return random.random() * (
                self.values[1] - self.values[0]
            ) + self.values[
                       0
                   ]  # type: ignore

        return random.choice(self.values)  # type: ignore

    def rand_outcomes(self, n: int, with_replacement=False, fail_if_not_enough=False) -> Iterable["Outcome"]:
        """Picks a random valid value."""
        if isinstance(self.values, int):
            if n > self.values and not with_replacement:
                if fail_if_not_enough:
                    raise ValueError(f'Cannot sample {n} outcomes out of {self.values} without replacement')
                else:
                    return [_ for _ in range(self.values)]
            if with_replacement:
                return np.random.randint(low=0, high=self.values, size=n).tolist()
            else:
                return random.shuffle([_ for _ in range(self.values)])[:n]
        elif self.is_continuous():
            if with_replacement:
                return (np.random.rand(n) * (self.values[1] - self.values[0]) + self.values[0]).tolist()
            else:
                return np.linspace(self.values[0], self.values[1], num=n, endpoint=True).tolist()

        if n > len(self.values) and not with_replacement:
            if fail_if_not_enough:
                raise ValueError(f'Cannot sample {n} outcomes out of {self.values} without replacement')
            else:
                return self.values
        return np.random.choice(np.asarray(self.values, dtype=type(self.values[0])), size=n
                                , replace=with_replacement).tolist()

    rand_valid = rand

    def rand_invalid(self) -> Union[int, float, str]:
        """Pick a random *invalid* value"""
        if isinstance(self.values, int):
            return random.randint(self.values + 1, 2 * self.values)

        elif self.is_continuous():
            return random.random() * (self.values[1]) + self.values[
                1
            ] + 1e-3  # type: ignore

        pick = unique_name('') + str(random.choice(self.values)) + unique_name(
            ''
        )  # type: ignore
        return pick

    @classmethod
    def enumerate(cls, issues: Collection['Issue'], max_n_outcomes: int=None, astype: Type = dict) -> List["Outcome"]:
        n = num_outcomes(issues)
        if n is None:
            return cls.sample(issues=issues, n_outcomes=max_n_outcomes, astype=astype)
        values = enumerate_outcomes(issues, keep_issue_names=False)
        if n > max_n_outcomes:
            values = random.sample(values, max_n_outcomes)
        outcomes = []
        for value in values:
            if astype == tuple:
                outcomes.append(tuple(value))
            elif astype == dict:
                outcomes.append(dict(zip((i.name for i in issues), value)))
            else:
                outcomes.append(astype(**dict(zip((i.name for i in issues), value))))
        return outcomes

    @classmethod
    def sample(cls, issues: Collection['Issue'], n_outcomes: int, astype: Type = dict
               , with_replacement: bool=True, fail_if_not_enough=True) -> List["Outcome"]:
        """
        Samples some outcomes from the issue space defined by the list of issues

        Args:

            issues: List of issues to sample from
            n_outcomes: The number of outcomes required
            astype: The type of the outcomes. It can be `tuple`, `dict` or any `OutcomeType`
            with_replacement: Whether sampling is with replacement (allowing repetition)
            fail_if_not_enough: IF given then an exception is raised if not enough outcomes are available

        Returns:

            a list of outcomes

        Examples:

            >>> from negmas import Issue, OutcomeType
            >>> issues = [Issue(name='price', values=(0.0, 3.0)), Issue(name='quantity', values=10)]

            Sampling outcomes as tuples

            >>> samples = Issue.sample(issues=issues, n_outcomes=10, astype=tuple)
            >>> len(samples) == 10
            True
            >>> type(samples[0]) == tuple
            True

            Sampling outcomes as dicts

            >>> samples = Issue.sample(issues=issues, n_outcomes=10, astype=dict)
            >>> len(samples) == 10
            True
            >>> type(samples[0]) == dict
            True
            >>> list(samples[0].keys())
            ['price', 'quantity']

            >>> @dataclass
            ... class MyOutcome(OutcomeType):
            ...     price: float = 0.0
            ...     quantity: int = 0


            Sampling outcomes as an arbitrary class

            >>> samples = Issue.sample(issues=issues, n_outcomes=10, astype=MyOutcome)
            >>> len(samples) == 10
            True
            >>> type(samples[0]) == MyOutcome
            True
            >>> list(samples[0].keys())
            ['price', 'quantity']


        """
        n_total = num_outcomes(issues)
        if n_total is not None and n_outcomes is not None and n_total < n_outcomes \
            and fail_if_not_enough and not with_replacement:
            raise ValueError(f'Cannot sample {n_outcomes} from a total of possible {n_total} outcomes')
        if n_total is not None and n_outcomes is None:
            values = enumerate_outcomes(issues=issues, keep_issue_names=False)
        elif n_total is None and n_outcomes is None:
            raise ValueError(f'Cannot sample unknown number of outcomes from continuous outcome spaces')
        else:
            samples = []
            for issue in issues:
                samples.append(issue.rand_outcomes(n=n_outcomes, with_replacement=True, fail_if_not_enough=True))
            values = []
            for i in range(n_outcomes):
                values.append([s[i] for s in samples])
            if not with_replacement:
                tmp_values = []
                for value in values:
                    tmp_values.append(tuple(value))
                tmp_values = set(tmp_values)
                remaining = n_outcomes - len(tmp_values)
                n_max_trials, i = 10, 0
                while remaining < 0 and i < n_max_trials:
                    tmp_values = tmp_values.union(set(cls.sample(issues=issues, n_outcomes=remaining, astype=tuple
                                                         , with_replacement=True, fail_if_not_enough=False)))
                    i += 1
                values = list(tmp_values)

        outcomes = []
        for value in values:
            if astype == tuple:
                outcomes.append(tuple(value))
            elif astype == dict:
                outcomes.append(dict(zip((i.name for i in issues), value)))
            else:
                outcomes.append(astype(**dict(zip((i.name for i in issues), value))))
        return outcomes

    @property
    def outcome_range(self) -> 'OutcomeRange':
        """An outcome range that represents the full space of the issues"""
        outcome_range = {}
        if self.is_continuous():
            outcome_range[self.name] = self.values
        else:
            outcome_range[self.name] = self.all
        return outcome_range

    def __str__(self):
        return f'{self.name}: {self.values}'

    __repr__ = __str__

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return self.values == other.values and self.name == other.name

    def __copy__(self):
        return Issue(name=self.name, values=self.values)

    def __deepcopy__(self, memodict={}):
        if isinstance(self.values, list):
            return Issue(name=self.name, values=[_ for _ in self.values])
        return Issue(name=self.name, values=self.values)

    @classmethod
    def from_java(cls, d: Dict[str, Any], class_name: str) -> 'Issue':
        if class_name.endswith('ListIssue'):
            return Issue(name=d.get('name', None), values=d['values'])
        if class_name.endswith('RangeIssue'):
            return Issue(name=d.get('name', None), values=(d['min'], d['max']))
        raise ValueError(f'Unknown issue type: {class_name} with dict {d} received from Java')

    def to_java(self):
        if self.values is None:
            return None
        if isinstance(self.values, tuple):
            if isinstance(self.values[0], int):
                return {'name': self.name, 'min': int(self.values[0]), 'max': int(self.values[0])
                        , PYTHON_CLASS_IDENTIFIER: 'negmas.outcomes.IntRangeIssue'}
            else:
                return {'name': self.name, 'min': float(self.values[0]), 'max': float(self.values[0])
                    , PYTHON_CLASS_IDENTIFIER: 'negmas.outcomes.DoubleRangeIssue'}
        if isinstance(self.values, Iterable):
            if isinstance(self.values[0], int):
                return {'name': self.name, 'values': [int(_) for _ in self.values]
                        , PYTHON_CLASS_IDENTIFIER: 'negmas.outcomes.IntListIssue'}
            elif isinstance(self.values[0], str):
                return {'name': self.name, 'values': [str(_) for _ in self.values]
                        , PYTHON_CLASS_IDENTIFIER: 'negmas.outcomes.StringListIssue'}
            return {'name': self.name, 'values': [float(_) for _ in self.values]
                    , PYTHON_CLASS_IDENTIFIER: 'negmas.outcomes.DoubleListIssue'}


class Issues(object):
    """Encodes a set of Issues.

    Args:

        name-value pairs



    Remarks:

        - Issues can be initialized by either an iterable of strings, an integer or a tuple of two real values with
          the following meanings:
          - ``iterable of strings``: This is an issue that can any value within the given set of values (strings)
          - ``int``: This is an issue that takes any value from 0 to the given value -1 (int)
          - ``float``: This is an issue that can take any real value between the given limits (min, max)

    """

    def __init__(self, **kwargs) -> None:
        self.issues = [Issue(name=k, values=v) for k, v in kwargs.items()]

    @classmethod
    def from_issue_collection(cls, issues: Iterable[Issue], name: str = None):
        out = Issues(name=name)
        out.issues = issues
        return out

    @classmethod
    def from_single_issue(cls, issue: Issue, name: str = None):
        return Issues.from_issue_collection([issue], name=name)

    def n_outcomes(self) -> int:
        """Returns the total number of outcomes in a set of issues. `-1` indicates infinity"""
        n = 1
        for issue in self.issues:
            n *= issue.cardinality()
            if n < 0:
                return -1

        return n

    @property
    def types(self) -> List[str]:
        """The type of the issue.

        Returns:
            str: either 'continuous' or 'discrete'

        """
        types_ = []
        for issue in self.issues:
            types_.append(issue.type)
        return types_

    def is_infinite(self) -> bool:
        """Test whether any issue is continuous (infinite outcome space)

        Returns:
            bool: continuous or not

        """
        return any(_.startswith('c') for _ in self.types)

    def is_finite(self) -> bool:
        """Test whether all issues are discrete (finite outcome space)

        Returns:
            bool: discrete or not

        """
        return all(_.startswith('d') for _ in self.types)

    @property
    def all(self) -> Generator:
        """A generator that generates all possible values.

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:


        """
        if self.is_infinite():
            raise ValueError(
                'Cannot return all possibilities of a continuous issues'
            )

        yield from itertools.product(_.all for _ in self.issues)

    def cardinality(self) -> int:
        """The number of possible outcomes for the issue. A negative number means infinite"""
        return self.n_outcomes()

    def rand(self) -> Dict[str, Union[int, float, str]]:
        """Picks a random valid value."""
        return {_.name: _.rand() for _ in self.issues}

    rand_valid = rand

    def rand_invalid(self) -> Dict[str, Union[int, float, str]]:
        """Pick a random *invalid* value"""
        return {_.name: _.rand_invalid() for _ in self.issues}

    @property
    def outcome_range(self) -> 'OutcomeRange':
        """An outcome range that represents the full space of the issues"""
        outcome_range = {}
        for issue in self.issues:
            if issue.is_continuous():
                outcome_range[issue.name] = issue.values
            else:
                outcome_range[issue.name] = issue.all
        return outcome_range

    def __str__(self):
        return '\n'.join([str(_) for _ in self.issues])


@dataclass
class OutcomeType:
    """A helper class allowing for definition of types that behave as outcomes (either in the form of dict or tuple).

    This class is intended to be used when a simple tuple or dict is not enough for describing an outcome (e.g. to use
    editor features like auto-completion of members). You simply define your class as a dataclass and add your fields to
    it then inherit from OutcomeType. As we do nothing in the __init__ function, that is compatible with python
    dataclasses.


    Examples:

        >>> from negmas import OutcomeType, Issue
        >>> @dataclass
        ... class MyOutcome(OutcomeType):
        ...     price: float = 0.0
        ...     quantity: int = 0

        You can use MyOutcome as an outcome directly or convert it to a tuple/dict for other functions

        >>> outcome = MyOutcome(price=2.0, quantity=3)
        >>> outcome.price
        2.0
        >>> outcome['price']
        2.0
        >>> outcome.astuple()
        (2.0, 3)
        >>> outcome.asdict()
        {'price': 2.0, 'quantity': 3}

        You can also use outputs from issues to initialize your class

        >>> issues = [Issue(name='price', values=(0.0, 3.0)), Issue(name='quantity', values=10)]
        >>> sample = Issue.sample(issues=issues, n_outcomes=1)[0]

        >>> outcome = MyOutcome(**sample)
        >>> outcome.price == outcome['price']
        True


    """

    def __getitem__(self, item):
        """Makes the outcome type behave like a dict"""
        return self.__dict__[item]

    def keys(self) -> List[str]:
        return tuple(_.name for _ in fields(self))

    def values(self) -> List[str]:
        return tuple(self.__dict__[_.name] for _ in fields(self))

    def astuple(self):
        """Converts the outcome to a tuple where the order of items is the same as they are defined as fields"""
        return tuple(self.__dict__[_.name] for _ in fields(self))

    def asdict(self):
        """Converts the outcome to a dict containing all fields"""
        return {_.name: self.__dict__[_.name] for _ in fields(self)}

    def get(self, name, default: Any = None):
        """Acts like dict.get"""
        try:
            return getattr(self, name, default)
        except:
            return default


Outcome = Union[OutcomeType, Tuple[Union[int, float, str, list]], Dict[Union[int, str], Union[int, float, str, list]]]
"""An outcome is either a tuple of values or a dict with name/value pairs."""

Outcomes = List['Outcome']
OutcomeRanges = List['OutcomeRange']


OutcomeRange = Mapping[
    Union[int, str],
    Union[
        int,
        float,
        str,
        List[int],
        List[float],
        List[str],
        Tuple[int, int],
        Tuple[float, float],
        List[Tuple[Union[int, float], Union[int, float]]],
    ],
]


def num_outcomes(issues: Collection[Issue]) -> Optional[int]:
    n = 1
    for issue in issues:
        c = issue.cardinality()
        if c < 0:
            return None
        n *= c
    return n


def enumerate_outcomes(issues: Iterable[Issue], keep_issue_names=True) \
    -> Optional[Union[List['Outcome'], Dict[str, 'Outcome']]]:
    """Enumerates all outcomes of this set of issues if possible

    Args:
        issues:
        keep_issue_names:

    Returns:
        List
    """
    try:
        outcomes = list(itertools.product(*[_.all for _ in issues]))
    except:
        return None
    if keep_issue_names:
        issue_names = [_.name for _ in issues]
        for i, outcome in enumerate(outcomes):
            outcomes[i] = dict(zip(issue_names, outcome))
    return outcomes


def sample_outcomes(issues: Iterable[Issue], n_outcomes: Optional[int] = None, keep_issue_names=True
                    , min_per_dim=5, expansion_policy=None) \
    -> Optional[List[Optional['Outcome']]]:
    """Discretizes the issue space and returns either a predefined number of outcomes or uniform samples

    Args:
        issues: The issues describing the issue space to be discretized
        n_outcomes: If None then exactly `min_per_dim` bins will be used for every continuous dimension and all outcomes
        will be returned
        keep_issue_names:
        min_per_dim:
        expansion_policy: None or 'repeat' or 'null' or 'no'. If repeat, then some of the outcomes will be repeated
        if None or 'no' then no expansion will happen if the total number of outcomes is less than
        n_outcomes. If 'null' then expansion will be with None values

    Returns:
        List


    Examples:

        enumberate the whole space
        >>> issues = [Issue(values=(0, 1), name='Price'), Issue(values=['a', 'b'], name='Name')]
        >>> sample_outcomes(issues=issues)
        [{'Price': 0.0, 'Name': 'a'}, {'Price': 0.0, 'Name': 'b'}, {'Price': 0.25, 'Name': 'a'}, {'Price': 0.25, 'Name': 'b'}, {'Price': 0.5, 'Name': 'a'}, {'Price': 0.5, 'Name': 'b'}, {'Price': 0.75, 'Name': 'a'}, {'Price': 0.75, 'Name': 'b'}, {'Price': 1.0, 'Name': 'a'}, {'Price': 1.0, 'Name': 'b'}]

        enumerate with sampling for very large space (we have 10 outcomes in the discretized space)
        >>> issues = [Issue(values=(0, 1), name='Price'), Issue(values=['a', 'b'], name='Name')]
        >>> sampled=sample_outcomes(issues=issues, n_outcomes=5)
        >>> len(sampled)
        5
        >>> len(set(tuple(_.values()) for _ in sampled))
        5


    """
    issues = [_ for _ in issues]
    continuous = []
    indx = []
    discrete = []
    n_disc = 0
    for i, issue in enumerate(issues):
        if issue.is_continuous():
            continuous.append(issue)
            indx.append(i)
        else:
            discrete.append(issue)
            n_disc += issue.cardinality()
    if len(continuous) > 0:
        if n_outcomes is not None:
            n_per_issue = max(min_per_dim, (n_outcomes - n_disc) / len(continuous))
        else:
            n_per_issue = min_per_dim
        for i, issue in enumerate(continuous):
            issues[indx[i]] = Issue(name=issue.name, values=list(np.linspace(issue.values[0], issue.values[1],
                                                                        num=n_per_issue, endpoint=True).tolist()))

    cardinality = 1
    for issue in issues:
        cardinality *= issue.cardinality()

    if cardinality == n_outcomes or n_outcomes is None:
        return list(enumerate_outcomes(issues, keep_issue_names=keep_issue_names))

    if cardinality < n_outcomes:
        outcomes = list(enumerate_outcomes(issues, keep_issue_names=keep_issue_names))
        if expansion_policy == 'no' or expansion_policy is None:
            return outcomes
        elif expansion_policy == 'null':
            return outcomes + [None] * (n_outcomes - cardinality)
        elif expansion_policy == 'repeat':
            n_reps = n_outcomes // cardinality
            n_rem = n_outcomes % cardinality
            if n_reps > 1:
                for _ in n_reps:
                    outcomes += outcomes
            if n_rem > 0:
                outcomes += outcomes[:n_rem]
            return outcomes

    return list(random.sample(enumerate_outcomes(issues, keep_issue_names=keep_issue_names), n_outcomes))


def _is_single(x):
    """Checks whether a value is a single value which is defined as either a string or not an Iterable."""
    return isinstance(x, str) or isinstance(x, int) or isinstance(x, float)


def outcome_is_valid(outcome: Outcome, issues: Collection[Issue]) -> bool:
    """Test validity of an outcome given a set of issues.

    Examples:
        >>> issues = [Issue((0.5, 2.0), 'price'), Issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')\
                , Issue(20, 'count')]
        >>> for _ in issues: print(_)
        price: (0.5, 2.0)
        date: ['2018.10.1', '2018.10.2', '2018.10.3']
        count: 20
        >>> print([outcome_is_valid({'price':3.0}, issues), outcome_is_valid({'date': '2018.10.4'}, issues)\
            , outcome_is_valid({'count': 21}, issues)])
        [False, False, False]
        >>> valid_incomplete = {'price': 1.9}
        >>> print(outcome_is_valid(valid_incomplete, issues))
        True
        >>> print(outcome_is_complete(valid_incomplete, issues))
        False
        >>> valid_incomplete.update({'date': '2018.10.2', 'count': 5})
        >>> print(outcome_is_complete(valid_incomplete, issues))
        True

    Args:
        outcome: outcome tested which can contain values for a partial set of issue values
        issues: issues

    Returns:
        Union[bool, Tuple[bool, str]]: If return_problem is True then a second return value contains a string with
                                      reason of failure
    """
    for issue in issues:
        for key in ikeys(outcome):
            if issue.name == str(key):
                break

        else:
            continue

        value = iget(outcome, key)
        if isinstance(issue.values, int) and (
            isinstance(value, str) or not 0 <= value < issue.values
        ):
            return False

        elif isinstance(issue.values, tuple) and (
            isinstance(value, str)
            or not issue.values[0] < value < issue.values[1]
        ):
            return False

        elif isinstance(issue.values, list) and value not in issue.values:
            return False

    return True


def outcome_is_complete(outcome: Outcome, issues: Collection[Issue]) -> bool:
    """Tests that the outcome is valid and complete.

    Examples:
        >>> issues = [Issue((0.5, 2.0), 'price'), Issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')\
                , Issue(20, 'count')]
        >>> for _ in issues: print(_)
        price: (0.5, 2.0)
        date: ['2018.10.1', '2018.10.2', '2018.10.3']
        count: 20
        >>> print([outcome_is_complete({'price':3.0}, issues), outcome_is_complete({'date': '2018.10.4'}, issues)\
            , outcome_is_complete({'count': 21}, issues)])
        [False, False, False]
        >>> valid_incomplete = {'price': 1.9}
        >>> print(outcome_is_complete(valid_incomplete, issues))
        False
        >>> valid_incomplete.update({'date': '2018.10.2', 'count': 5})
        >>> print(outcome_is_complete(valid_incomplete, issues))
        True
        >>> invalid = {'price': 2000, 'date': '2018.10.2', 'count': 5}
        >>> print(outcome_is_complete(invalid, issues))
        False
        >>> invalid = {'unknown': 2000, 'date': '2018.10.2', 'count': 5}
        >>> print(outcome_is_complete(invalid, issues))
        False

     Args:
        outcome: outcome tested which much contain valid values all issues if it is to be considered complete.
        issues: issues

    Returns:
        Union[bool, Tuple[bool, str]]: If return_problem is True then a second return value contains a string with
                                      reason of failure

    """
    if len(outcome) != len(issues):
        return False

    valid = outcome_is_valid(outcome, issues)
    if not valid:
        return False

    for issue in issues:
        if issue.name not in outcome.keys():
            return False

    return True


def outcome_range_is_valid(
    outcome_range: OutcomeRange, issues: Optional[Collection[Issue]] = None
) -> Union[bool, Tuple[bool, str]]:
    """Tests whether the outcome range is valid for the set of issues.

    Args:
        outcome_range:
        issues:

    Example:
        >>> try:
        ...     outcome_range_is_valid({'price': (0, 10)})
        ... except NotImplementedError:
        ...     print('Not implemented')
        Not implemented

    Returns:

    """
    # TODO implement this function
    raise NotImplementedError()


def outcome_range_is_complete(
    outcome_range: OutcomeRange, issues: Optional[Collection[Issue]] = None
) -> Union[bool, Tuple[bool, str]]:
    """Tests whether the outcome range is valid and complete for the set of issues

    Args:
        outcome_range:
        issues:

    Example:
        >>> try:
        ...     outcome_range_is_complete({'price': (0, 10)})
        ... except NotImplementedError:
        ...     print('Not implemented')
        Not implemented

    Returns:

    """
    # TODO implement this function
    raise NotImplementedError()


#################################
# Outcome space implementation  #
#################################
def outcome_in_range(
    outcome: Outcome,
    outcome_range: OutcomeRange,
    *,
    strict=False,
    fail_incomplete=False,
) -> bool:
    """
    Tests that the outcome is contained within the given range of outcomes.

    An outcome range defines a value or a range of values for each issue.

    Args:

        outcome: Outcome being tested
        outcome_range: Outcome range being tested against
        strict: Whether to enforce that all issues in the outcome must be mentioned in the outcome_range
        fail_incomplete: If True then outcomes that do not sepcify a value for all keys in the outcome_range
        will be considered not falling within it. If False then these outcomes will be considered falling
        within the range given that the values for the issues mentioned in the outcome satisfy the range
        constraints.

    Examples:

        >>> outcome_range = {'price': (0.0, 2.0), 'distance': [0.3, 0.4], 'type': ['a', 'b'], 'area': 3}
        >>> outcome_range_2 = {'price': [(0.0, 1.0), (1.5, 2.0)], 'area': [(3, 4), (7, 9)]}
        >>> outcome_in_range({'price':3.0}, outcome_range)
        False
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range)
        True
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range, strict=True)
        False
        >>> outcome_in_range({'area': 3}, outcome_range, fail_incomplete=True)
        False
        >>> outcome_in_range({'area': 3}, outcome_range)
        True
        >>> outcome_in_range({'type': 'c'}, outcome_range)
        False
        >>> outcome_in_range({'type': 'a'}, outcome_range)
        True
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range_2)
        True
        >>> outcome_in_range({'area': 3.1}, outcome_range_2)
        True
        >>> outcome_in_range({'area': 3}, outcome_range_2)
        False
        >>> outcome_in_range({'area': 5}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 0.4}, outcome_range_2)
        True
        >>> outcome_in_range({'price': 0.4}, outcome_range_2, fail_incomplete=True)
        False
        >>> outcome_in_range({'price': 1.2}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 0.4, 'area': 3.9}, outcome_range_2)
        True
        >>> outcome_in_range({'price': 0.4, 'area': 10}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 1.2, 'area': 10}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 1.2, 'area': 4}, outcome_range_2)
        False
        >>> outcome_in_range({'type': 'a'}, outcome_range_2)
        True
        >>> outcome_in_range({'type': 'a'}, outcome_range_2, strict=True)
        False
        >>> outcome_range = {'price': 10}
        >>> outcome_in_range({'price': 10}, outcome_range)
        True
        >>> outcome_in_range({'price': 11}, outcome_range)
        False

    Returns:

        bool: Success or failure

    Remarks:
        Outcome ranges specify regions in an outcome space. They can have any of the following conditions:

        - A key/issue not mentioned in the outcome range does not add any constraints meaning that **All**
          values are acceptable except if strict == True. If strict == True then *NO* value will be accepted for issues
          not in the outcome_range.
        - A key/issue with the value None in the outcome range means **All** values on this issue are acceptable.
          This is the same as having this key/issue removed from the outcome space
        - A key/issue withe the value [] (empty list) accepts *NO* outcomes
        - A key/issue with  a single value means that it is the only one acceptable
        - A key/issue with a single 2-items tuple (min, max) means that any value within that range is acceptable.
        - A key/issue with a list of values means an output is acceptable if it falls within the condition specified
          by any of the values in the list (list == union). Each such value can be a single value, a 2-items
          tuple or another list. Notice that lists of lists can always be combined into a single list of values

    """
    if fail_incomplete and len(
        set(ikeys(outcome_range)).difference(ikeys(outcome))
    ) > 0:
        return False

    for key, value in ienumerate(outcome):
        if key not in ikeys(outcome_range):
            if strict:
                return False

            continue

        values = iget(outcome_range, key, None)
        if values is None:
            return False

        if _is_single(values) and value != values:
            return False

        if isinstance(values, tuple) and not values[0] < value < values[1]:
            return False

        if isinstance(values, list):
            for constraint in values:
                if _is_single(constraint):
                    if value == constraint:
                        break

                elif isinstance(constraint, list):
                    if value in constraint:
                        break

                elif isinstance(constraint, tuple):
                    if constraint[0] < value < constraint[1]:
                        break

            else:
                return False

            continue

    return True

#
# def _make_iterable(
#     v: Union[list, tuple, int, str, float]
# ) -> Union[list, tuple]:
#     """Ensures the output is an Iterable if it was a single value.
#
#     >>> [_make_iterable(1), _make_iterable('1'), _make_iterable(1.0), _make_iterable([1])]
#     [{1}, {'1'}, {1.0}, {1}]
#
#     """
#     if _is_single(v):
#         return {v}
#
#     return set(v)
#
#
# def _intersect_tuples(v1: tuple, v2: tuple) -> tuple:
#     """Finds intersection between two tuples.
#
#     Examples:
#
#         >>> print(_intersect_tuples((0.0, 1.0), (1.1, 2.0)))
#         (0.0, 0.0)
#         >>> _intersect_tuples((0.0, 1.0), (1.0, 2.0))
#         (1.0, 1.0)
#         >>> _intersect_tuples((0.0, 1.0), (0.5, 2.0))
#         (0.5, 1.0)
#
#     """
#     if not (v1[0] <= v2[0] <= v1[1] or v2[0] <= v1[0] <= v2[1]):
#         return (v1[0], v1[0])
#
#     return max((v1[0], v2[0])), min((v1[1], v2[1]))
#
#
# def _union_tuples(v1: tuple, v2: tuple) -> Union[tuple, List[tuple]]:
#     """The union of two tuples.
#
#     Examples:
#
#         >>> _union_tuples((0.0, 1.0), (0.0, 1.0))
#         (0.0, 1.0)
#         >>> _union_tuples((0.0, 1.0), (0.5, 1.0))
#         (0.0, 1.0)
#         >>> _union_tuples((0.0, 1.0), (0.5, 2.0))
#         (0.0, 2.0)
#         >>> _union_tuples((0.0, 1.0), (1.0, 2.0))
#         [(0.0, 1.0), (1.0, 2.0)]
#         >>> _union_tuples((0.0, 1.0), (2.0, 3.0))
#         [(0.0, 1.0), (2.0, 3.0)]
#
#     """
#     if v1[0] == v2[0] and v1[1] == v2[1]:
#         return v1
#
#     if not (v1[0] < v2[0] < v1[1] or v2[0] < v1[0] < v2[1]):
#         return [v1, v2]
#
#     return min((v1[0], v2[0])), max((v1[1], v2[1]))
#
#
# def _subtract_tuples(v1: tuple, v2: tuple) -> Union[tuple, List[tuple]]:
#     """The difference between first and second tuple (v1 - v2)
#
#     Examples:
#
#         >>> _subtract_tuples((0.0, 1.0), (-1.0, 0.0))
#         (0.0, 1.0)
#         >>> _subtract_tuples((0.0, 1.0), (-1.0, 0.5))
#         (0.5, 1.0)
#         >>> _subtract_tuples((0.0, 1.0), (0.2, 0.5))
#         [(0.0, 0.2), (0.5, 1.0)]
#         >>> _subtract_tuples((0.0, 1.0), (0.5, 1.5))
#         (0.0, 0.5)
#         >>> _subtract_tuples((0.0, 1.0), (1.5, 2.0))
#         (0.0, 1.0)
#
#     """
#     v = _intersect_tuples(v1, v2)
#     if v[0] == v[1]:
#         return v1
#
#     if v[0] <= v1[0] and v[1] <= v1[1]:
#         return (v[1], v1[1])
#
#     if v[0] > v1[0] and v[1] < v1[1]:
#         return [(v1[0], v[0]), (v[1], v1[1])]
#
#     return (v1[0], v[0])
#
#
# def _apply_on_lists(
#     v1: Union[set, list], v2: Union[set, list], op: str
# ) -> set:
#     """Applies a set operation on lists.
#
#     Examples:
#
#         >>> _apply_on_lists([1, 2, 3], [2, 4 ,5], 'intersection')
#         {2}
#         >>> _apply_on_lists([1, 2, 3], [2, 4 ,5], 'union')
#         {1, 2, 3, 4, 5}
#         >>> _apply_on_lists([1, 2, 3], [2, 4 ,5], 'subtract')
#         {1, 3}
#
#     """
#     v1, v2 = set(v1), set(v2)
#     if op.startswith('i'):
#         return v1.intersection(v2)
#
#     elif op.startswith('u'):
#         return v1.union(v2)
#
#     elif op.startswith('s'):
#         return v1.difference(v2)
#     raise ValueError()
#
#
# def _intersect_list_tuple(lst: list, tpl: tuple) -> list:
#     """Intersection between a list and tuple.
#
#     Examples:
#
#         >>> _intersect_list_tuple([1, 2, 3], (2.5, 4.0))
#         [3]
#         >>> _intersect_list_tuple([1, 2, 3], (6.0, 7.0))
#         []
#         >>> _intersect_list_tuple([1, 2, 3], (0.0, 3.0))
#         [1, 2, 3]
#         >>> _intersect_list_tuple([], (0.0, 3.0))
#         []
#
#     """
#     result = []
#     for _ in lst:
#         if tpl[0] <= _ <= tpl[1]:
#             result.append(_)
#     return result
#
#
# def _union_list_tuple(
#     lst: list, tpl: tuple
# ) -> Union[tuple, List[Union[tuple, list]]]:
#     """Finds the union between a list and a tuple.
#
#     Examples:
#
#         >>> _union_list_tuple([1, 2, 3], (2.5, 4.0))
#         [(2.5, 4.0), [1, 2, 3]]
#         >>> _union_list_tuple([1, 2, 3], (6.0, 7.0))
#         [(6.0, 7.0), [1, 2, 3]]
#         >>> _union_list_tuple([1, 2, 3], (0.0, 3.0))
#         (0.0, 3.0)
#         >>> _union_list_tuple([], (0.0, 3.0))
#         (0.0, 3.0)
#
#     """
#     for _ in lst:
#         if not (tpl[0] <= _ <= tpl[1]):
#             return [tpl, lst]
#
#     return tpl


# def range_intersection(r1: OutcomeRange, r2: OutcomeRange) -> OutcomeRange:
#     """A new outcome range at the intersection of given outcome ranges
#
#     Examples:
#         >>> r1 = {'price': (0.0, 1.0), 'cost':[1, 2, 3, 4, 5]
#         ...        , 'delivery': ['yes', 'no'], 'honor':'yes'}
#         >>> r2 = {'price': (0.5, 2.0), 'cost':[4, 5, 6]
#         ...        , 'delivery': ['yes'], 'honor':'yes'}
#         >>> r3 = {'price': (0.5, 2.0), 'cost':[4, 5, 6]
#         ...        , 'delivery': ['yes'], 'honor':'no'}
#         >>> r4 = {'price': (0.5, 2.0), 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':'yes'}
#         >>> r5 = {'price': [1.0, 3.0], 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':None}
#         >>> r6 = {'price': [1.0, 3.0], 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':[]}
#         >>> r7 = {'price': 9.0, 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':None}
#         >>> r8 = {'price': 0.5, 'cost':[9]
#         ...        , 'delivery': 'no', 'honor':None}
#         >>> r9 = {'price': 0.5}
#         >>> range_intersection(r1, r2)
#         {'price': (0.5, 1.0), 'cost': [4, 5], 'delivery': 'yes', 'honor': 'yes'}
#         >>> range_intersection(r1, r3)
#         {'price': (0.5, 1.0), 'cost': [4, 5], 'delivery': 'yes', 'honor': []}
#         >>> range_intersection(r1, r4)
#         {'price': (0.5, 1.0), 'cost': [4, 5], 'delivery': 'yes', 'honor': 'yes'}
#         >>> range_intersection(r1, r5)
#         {'price': 1.0, 'cost': [4, 5], 'delivery': 'yes', 'honor': 'yes'}
#         >>> range_intersection(r1, r6)
#         {'price': 1.0, 'cost': [4, 5], 'delivery': 'yes', 'honor': []}
#         >>> range_intersection(r1, r7)
#         {'price': [], 'cost': [4, 5], 'delivery': 'yes', 'honor': 'yes'}
#         >>> range_intersection(r1, r8)
#         {'price': 0.5, 'cost': [], 'delivery': 'no', 'honor': 'yes'}
#         >>> range_intersection(r1, r9)
#         {'price': 0.5, 'cost': [1, 2, 3, 4, 5], 'delivery': ['yes', 'no'], 'honor': 'yes'}
#
#     """
#     r = {}
#     r1 = range_simplify(r1, compress=False)
#     r2 = range_simplify(r2, compress=False)
#     for k, v1 in r1.items():
#         if k not in r2.keys():
#             r[k] = copy.copy(v1)
#             continue
#         v2 = r2[k]
#         if v1 is None:
#             r[k] = copy.copy(v2)
#             continue
#         if v2 is None:
#             r[k] = copy.copy(v1)
#             continue
#         v1 = copy.copy(_make_iterable(v1))
#         v2 = copy.copy(_make_iterable(v2))
#         if isinstance(v1, tuple) and isinstance(v2, tuple):
#             r[k] = _intersect_tuples(v1, v2)
#         elif isinstance(v1, tuple) and not isinstance(v2, tuple):
#             r[k] = _intersect_list_tuple(lst=v2, tpl=v1)
#         elif not isinstance(v1, tuple) and isinstance(v2, tuple):
#             r[k] = _intersect_list_tuple(lst=v1, tpl=v2)
#         else:
#             r[k] = set(v1).intersection(v2)
#         if isinstance(r[k], list) and len(r[k]) == 1:
#             r[k] = r[k][0]
#     return r
#
#
# def range_simplify(r: OutcomeRange, compress: bool=False) -> OutcomeRange:
#     """Simplifies an outcome range.
#
#     Args:
#
#         r: The outcome range
#         compress: If true the dimensions that are not constrained are removed.
#
#     Examples:
#
#         >>> range_simplify({'a': [5], 'b': [4, 1, 3, 2], 'c': (1.0, 1.0)})
#         {'a': 5, 'b': [4, 1, 3, 2], 'c': 1.0}
#
#         >>> range_simplify({'a': [5], 'b': [4, 1, 3, 2], 'c': [(0.0, 2.0), (1.5, 3.0)
#         ...                              , 4, 2, [1, 2, 3], [1], (0.5, 0.5)]})
#         {'a': 5, 'b': [1, 2, 3, 4], 'c': [(0.0, 3.0), [1, 2, 3, 4]]
#
#
#     """
#
#     def _simplify_value(cutoff_utility):
#         """Puts the value in the most compressed form. """
#
#         if _is_single(cutoff_utility):
#             return cutoff_utility
#         if isinstance(cutoff_utility, tuple) and cutoff_utility[0] == cutoff_utility[1]:
#             return cutoff_utility[0]
#         if isinstance(cutoff_utility, list) and len(cutoff_utility) == 1:
#             return cutoff_utility
#         if isinstance(cutoff_utility, list):
#             combined_list = []
#             tuples = []
#             for item in cutoff_utility:
#                 if _is_single(item):
#                     combined_list.append(item)
#                 elif isinstance(item, tuple):
#                     tuples.append(item)
#                 elif isinstance(item, list):
#                     combined_list += item
#                 else:
#                     raise ValueError('not a single, tuple or list!!! cannot understand')
#             if len(tuples) > 1:
#                 results = _union_tuples(tuples[0], tuples[1])
#                 for i in range(2, len(tuples)):
#                     results = _union_tuples(results, tuples[i])
#                 tuples = results
#             if isinstance(tuples, tuple):
#                 tuples = [tuples]
#             cutoff_utility = tuples + sorted(list(set(combined_list)))
#             if len(cutoff_utility) == 1:
#                 cutoff_utility = cutoff_utility[0]
#         return cutoff_utility
#
#     keys = sorted(r.keys())
#     return {k: _simplify_value(r[k]) for k in keys}
#
#
# def range_equal(r1: OutcomeRange, r2: OutcomeRange) -> OutcomeRange:
#     """Tests if the two outcomes are equal
#
#     Examples:
#         >>> range_equal({'a': 5, 'b': [1, 2, 3, 4], 'c': (1.0, 2.0)},
#         ...             {'a': [5], 'b': [4, 1, 3, 2], 'c': (1.0, 2.0)})
#
#     """
#
# def range_union(r1: OutcomeRange, r2: OutcomeRange) -> OutcomeRange:
#     """A new outcome range covering the union of input outcome ranges.
#
#     Examples:
#         >>> r1 = {'price': (0.0, 1.0), 'cost':[1, 2, 3, 4, 5]
#         ...        , 'delivery': ['yes', 'no'], 'honor':'yes'}
#         >>> r2 = {'price': (0.5, 2.0), 'cost':[4, 5, 6]
#         ...        , 'delivery': ['yes'], 'honor':'yes'}
#         >>> r3 = {'price': (0.5, 2.0), 'cost':[4, 5, 6]
#         ...        , 'delivery': ['yes'], 'honor':'no'}
#         >>> r4 = {'price': (0.5, 2.0), 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':'yes'}
#         >>> r5 = {'price': [1.0, 3.0], 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':None}
#         >>> r6 = {'price': [1.0, 3.0], 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':[]}
#         >>> r7 = {'price': 9.0, 'cost':[4, 5, 6]
#         ...        , 'delivery': 'yes', 'honor':None}
#         >>> r8 = {'price': 0.5, 'cost':[9]
#         ...        , 'delivery': 'no', 'honor':None}
#         >>> r9 = {'price': 0.5}
#         >>> range_union(r1, r2)
#         {'honor': 'yes', 'cost': [1, 2, 3, 4, 5, 6], 'price': (0.0, 2.0), 'delivery': ['yes', 'no']}
#         >>> range_union(r1, r3)
#         {'honor': ['yes', 'no'], 'cost': [1, 2, 3, 4, 5, 6], 'price': (0.0, 2.0), 'delivery': ['yes', 'no']}
#         >>> range_union(r1, r4)
#         {'honor': 'yes', 'cost': [1, 2, 3, 4, 5, 6], 'price': (0.0, 2.0), 'delivery': ['yes', 'no']}
#         >>> range_union(r1, r5)
#         {'honor': None, 'cost': [1, 2, 3, 4, 5, 6], 'price': [(0.0, 1.0), [3]], 'delivery': ['yes', 'no']}
#         >>> range_union(r1, r6)
#         {'honor': 'yes', 'cost': [1, 2, 3, 4, 5, 6], 'price': [(0.0, 1.0), [3]], 'delivery': ['yes', 'no']}
#         >>> range_union(r1, r7)
#         {'honor': None, 'cost': [1, 2, 3, 4, 5, 9], 'price': [(0.0, 1.0), [9.0]], 'delivery': ['yes', 'no']}
#         >>> range_union(r1, r8)
#         {'honor': None, 'cost': [1, 2, 3, 4, 5, 6], 'price': (0.0, 1.0), 'delivery': ['yes', 'no']}
#         >>> range_union(r1, r9)
#         {'honor': None, 'cost': None, 'price': (0.0, 1.0), 'delivery': None}
#
#     """
#     r = {}
#     all_k = list(set(r1.keys()).union(r2.keys()))
#     for k in all_k:
#         if k not in r1.keys():
#             r[k] = None
#             continue
#         if k not in r2.keys():
#             r[k] = None
#             continue
#         v1, v2 = r1[k], r2[k]
#         if v1 is None or v2 is None:
#             r[k] = None
#             continue
#         v1 = _make_iterable(v1)
#         v2 = _make_iterable(v2)
#         if isinstance(v1, tuple) and isinstance(v2, tuple):
#             r[k] = _union_tuples(v1, v2)
#         elif isinstance(v1, tuple) and not isinstance(v2, tuple):
#             r[k] = _union_list_tuple(lst=v2, tpl=v1)
#         elif not isinstance(v1, tuple) and isinstance(v2, tuple):
#             r[k] = _union_list_tuple(lst=v1, tpl=v2)
#         else:
#             r[k] = _apply_on_lists(v1, v2, 'union')
#     return r
#
# def range_subtract(r1: OutcomeRange, r2: OutcomeRange, parent1: OutcomeRange=None, parent2: OutcomeRange=None) -> OutcomeRange:
#     r = {}
#
#     for k, v1 in r1.items():
#         if k not in r2.keys():
#             r[k] = []
#             continue
#         v2 = r2[k]
#         if v1 is None:
#             if parent1 is None:
#                 raise RuntimeError('No parent is defined, cannot invert')
#             r[k] = range_invert({k: v2}, {k: parent1.get(k, None)})
#             continue
#         v1 = _make_iterable(v1)
#         v2 = _make_iterable(v2)
#         if isinstance(v1, tuple) and isinstance(v2, tuple):
#             r[k] = _subtract_tuples(v1, v2)
#         elif isinstance(v1, tuple) and not isinstance(v2, tuple):
#             r[k] = _subtract_list_tuple(lst=v2, tpl=v1)
#         elif not isinstance(v1, tuple) and isinstance(v2, tuple):
#             r[k] = _subtract_list_tuple(lst=v1, tpl=v2)
#         else:
#             r[k] = _apply_on_lists(v1, v2, 'subtract')
#     return r
#
#
# def range_invert(r1: OutcomeRange, issues=Iterable[Issue], resolution=1e-6) -> OutcomeRange:
#     r = {}
#
#     for k, v1 in r1.items():
#         if v1 is None:
#             r[k] = []
#             continue
#     # TODO complete this
#
#     return r
#
#
# def _dim_cardinality(x: Any, recursing=False) -> Union[int, float]:
#     if isinstance(x, str):
#         return 1
#     if isinstance(x, int) and not recursing:
#         return x
#     if isinstance(x, tuple):
#         if x[0] != x[1]:
#             return math.inf
#         else:
#             return 1
#     if isinstance(x, Iterable):
#         return sum([_dim_cardinality(_, recursing=True) for _ in x])
#     return 1
#
#
# def _all_values(x: Any, recursing=False):
#     if isinstance(x, str):
#         yield x
#         raise StopIteration
#     if isinstance(x, int) and not recursing:
#         yield from range(x)
#         raise StopIteration
#     if isinstance(x, tuple):
#         raise RuntimeError('Cannot iterate over ranges (tuples)')
#     if isinstance(x, Iterable):
#         yield from x
#     raise StopIteration
SingleValue = Union[int, float, str]
EPSILON = 1e-9


# class SingleIssueSpace(NamedObject):
#     """Encodes the space of values for a single issue.
#
#     Examples:
#
#         You can define a complex issue space combining ranges and sets of values:
#         >>> everything = SingleIssueSpace(ranges= (20.0, 30.0), count = 10, name='everything')
#         >>> print(everything)
#         everything: (20.0, 30.0) and {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}
#
#         You can create child issue spaces that belong to it as needed
#         >>> i1 = SingleIssueSpace(ranges= (20.0, 25.0), values = [1, 2, 3, 4, 5], name='i1', parent=everything)
#         >>> i2 = SingleIssueSpace(ranges= (20.0, 25.0), name='i2', parent=everything)
#         >>> i3 = SingleIssueSpace(values = [1, 2, 3, 4, 5], name='i3', parent=everything)
#         >>> i4 = SingleIssueSpace(values = [6, 7, 8, 9], ranges = (25.0, 29.0), name='i3', parent=everything)
#
#         You can make complex spaces by combining them arbitrary using arithmetic operators
#         >>> print(everything - ((i2 + i3) ))
#         (everything-(i2+i3)): (25.0, 30.0) and {0, 6, 7, 8, 9}
#         >>> print(-(i2 + i4))
#         -(i2+i3): (29.0, 30.0) and {0, 1, 2, 3, 4, 5}
#
#         You can also use logical operators
#         >>> print(everything % ((i2 | i3) ))
#         (everything-(i2+i3)): (25.0, 30.0) and {0, 6, 7, 8, 9}
#         >>> print(~(i2 | i4))
#         -(i2+i3): (29.0, 30.0) and {0, 1, 2, 3, 4, 5}
#
#
#         See the documentation of arithmetic operators to find more examples
#
#     """
#     __slots__ = ['ranges', 'values', 'parent', 'everything']
#
#     def __init__(
#         self,
#         ranges: Union[Tuple[Any, Any], Iterable[Tuple[Any, Any]]] = None,
#         values: Iterable[SingleValue] = None,
#         count: int = None,
#         parent: 'SingleIssueSpace' = None,
#         everything=False,
#         name: str = None,
#     ):
#         super().__init__(name=name)
#         self.parent = parent
#         self.everything = everything
#         if everything:
#             self.ranges = self.values = None
#         else:
#             if ranges is None:
#                 ranges = []
#             if values is None:
#                 values = []
#             if isinstance(ranges, tuple):
#                 ranges = [ranges]
#             self.ranges = list(ranges)
#             self.values = set(values)
#             if count is not None:
#                 self.values = self.values.union(range(count))
#             self.simplify()
#
#     @property
#     def n_outcomes(self) -> Union[int, float]:
#         """Returns the total number of outcomes in a set of issues. `math.inf` indicates infinity
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> SingleIssueSpace(ranges=(0.0, 1.0), values=[2.0, 3.0, 4.0]).n_outcomes
#             inf
#             >>> SingleIssueSpace(values=[2.0, 3.0, 4.0]).n_outcomes
#             3
#             >>> SingleIssueSpace(count=10).n_outcomes
#             10
#             >>> SingleIssueSpace(count=10, values=[20, 30, 40]).n_outcomes
#             13
#             >>> SingleIssueSpace(count=10, values=[20, 30, 40, 40, 40]).n_outcomes
#             13
#             >>> SingleIssueSpace(count=10, values=[1, 2, 4]).n_outcomes
#             10
#
#
#         """
#         if self.everything:
#             return self.parent.n_outcomes if self.parent is not None else math.inf
#
#         if len(self.ranges) > 0:
#             return math.inf
#
#         return len(self.values)
#
#     @property
#     def cardinalities(self) -> Dict[str, Union[int, float]]:
#         """The cardinality of that issue space with its name. Useful when creating complex OutcomeSpaces
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> SingleIssueSpace(name='issue', ranges=(0.0, 1.0), values=[2.0, 3.0, 4.0]).cardinalities
#             {'issue': inf}
#             >>> SingleIssueSpace(name='issue', values=[2.0, 3.0, 4.0]).cardinalities
#             {'issue': 3}
#             >>> SingleIssueSpace(name='issue', count=10).cardinalities
#             {'issue': 10}
#             >>> SingleIssueSpace(name='issue', count=10, values=[20, 30, 40]).cardinalities
#             {'issue': 13}
#             >>> SingleIssueSpace(name='issue', count=10, values=[20, 30, 40, 40, 40]).cardinalities
#             {'issue': 13}
#             >>> SingleIssueSpace(name='issue', count=10, values=[1, 2, 4]).cardinalities
#             {'issue': 10}
#
#         """
#         return {self.name: self.cardinality}
#
#     @property
#     def type(self) -> str:
#         """The type of the issue.
#
#         Returns:
#             str: either 'continuous' or 'discrete'
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> SingleIssueSpace(name='issue', ranges=(0.0, 1.0), values=[2.0, 3.0, 4.0]).type
#             'continuous'
#             >>> SingleIssueSpace(name='issue', values=[2.0, 3.0, 4.0]).type
#             'discrete'
#             >>> SingleIssueSpace(name='issue', count=10).type
#             'discrete'
#             >>> SingleIssueSpace(name='issue', count=10, values=[20, 30, 40]).type
#             'discrete'
#
#         """
#         if self.everything:
#             return self.parent.type if self.parent is not None else (
#                 'continuous' if len(self.ranges) > 0 else 'discrete'
#             )
#
#         return 'continuous' if len(self.ranges) > 0 else 'discrete'
#
#     @property
#     def infinite(self) -> bool:
#         """Test whether any issue is continuous (infinite outcome space)
#
#         Returns:
#             bool: continuous (== infinite) or not
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> SingleIssueSpace(name='issue', ranges=(0.0, 1.0), values=[2.0, 3.0, 4.0]).infinite
#             True
#             >>> SingleIssueSpace(name='issue', values=[2.0, 3.0, 4.0]).infinite
#             False
#             >>> SingleIssueSpace(name='issue', count=10).infinite
#             False
#             >>> SingleIssueSpace(name='issue', count=10, values=[20, 30, 40]).infinite
#             False
#
#         """
#         if self.everything:
#             return self.parent.infinite if self.parent is not None else len(
#                 self.ranges
#             ) > 0
#
#         return len(self.ranges) > 0
#
#     @property
#     def finite(self) -> bool:
#         """Test whether all issues are discrete (finite outcome space)
#
#         Returns:
#             bool: discrete or not
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> SingleIssueSpace(name='issue', ranges=(0.0, 1.0), values=[2.0, 3.0, 4.0]).finite
#             False
#             >>> SingleIssueSpace(name='issue', values=[2.0, 3.0, 4.0]).finite
#             True
#             >>> SingleIssueSpace(name='issue', count=10).finite
#             True
#             >>> SingleIssueSpace(name='issue', count=10, values=[20, 30, 40]).finite
#             True
#
#         """
#         return not self.infinite
#
#     @property
#     def all(self) -> Generator:
#         """A generator that generates all possible values.
#
#         Remarks:
#             - This function returns a generator for the case when the number of values is very large.
#             - If you need a list then use something like:
#
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> try:
#             ...     list(SingleIssueSpace(name='issue', ranges=(0.0, 1.0), values=[2.0, 3.0, 4.0]).all)
#             ... except ValueError:
#             ...     print('failed')
#             failed
#             >>> list(SingleIssueSpace(name='issue', values=[2.0, 3.0, 4.0]).all)
#             [2.0, 3.0, 4.0]
#             >>> list(SingleIssueSpace(name='issue', count=10).all)
#             [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
#             >>> parent = SingleIssueSpace(name='issue', count=10)
#             >>> list(SingleIssueSpace(name='issue', count=3, parent=parent).all)
#             [0, 1, 2]
#             >>> list(SingleIssueSpace(name='issue', parent=parent).all)
#             []
#             >>> list(SingleIssueSpace(name='issue', everything=True, parent=parent).all)
#             [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
#
#         """
#         if self.infinite:
#             raise ValueError(
#                 'Cannot return all possibilities of a continuous issues'
#             )
#
#         if self.everything:
#             yield from self.parent.all
#
#         else:
#             yield from self.values
#
#     def __contains__(self, item: SingleValue) -> bool:
#         """Tests if an outcome is within this issue space.
#
#
#         Examples:
#
#             >>> prices = SingleIssueSpace(name='price', ranges = [(0.0, 10.0), (20.0, 30.0)], values=[2.0, 3.0, 44])
#             >>> [_ in prices for _ in (0.0, 0.5, 1.0, 15.0, 20.0, 22.0, 40, 44.0)]
#             [False, True, True, False, False, True, False, True]
#
#         """
#         if self.everything:
#             return item in self.parent
#
#         for r in self.ranges:
#             if r[0] < item < r[1]:
#                 return True
#
#         return item in self.values
#
#     def _in_ranges(self, v: SingleValue):
#         """Checks if a value is within the ranges."""
#         for r in self.ranges:
#             if r[0] < v < r[1]:
#                 return True
#
#         return False
#
#     def _compress_ranges(self) -> None:
#         """Combines ranges and removes empty ones."""
#         self.ranges = sorted(
#             [_ for _ in self.ranges if _[0] != _[1]], key=lambda x: x[0]
#         )
#         if len(self.ranges) > 1:
#             ranges = _union_tuples(self.ranges[0], self.ranges[1])
#             for i in range(2, len(self.ranges)):
#                 if isinstance(ranges, tuple):
#                     ranges = _union_tuples(ranges, self.ranges[i])
#                 else:
#                     last_union = _union_tuples(ranges[-1], self.ranges[i])
#                     if isinstance(last_union, tuple):
#                         last_union = [last_union]
#                     ranges = ranges[:-1] + last_union
#             if isinstance(ranges, tuple):
#                 ranges = [ranges]
#             self.ranges = ranges
#
#     def simplify(self) -> None:
#         """Simplifies internal representation.
#
#         Examples:
#
#             >>> print(SingleIssueSpace(count=8, ranges= [(22.0, 34.0), (25, 41), (5.0, 7.0)]
#             ...                                     , values= [12, 11, 10, 13, 50], name='i2'))
#             i2: [(5.0, 7.0), (22.0, 41)] and {0, 1, 2, 3, 4, 5, 7, 10, 11, 12, 13, 50}
#             >>> print(SingleIssueSpace(ranges=[(0.0, 1.0), (2.0, 3.0)], values={10.0, 11.0}, name='issue 1'))
#             issue 1: [(0.0, 1.0), (2.0, 3.0)] and {10.0, 11.0}
#             >>> print(SingleIssueSpace(ranges=[(0.0, 1.0), (0.5, 3.0)], values={10.0, 11.0}, name='issue 2'))
#             issue 2: (0.0, 3.0) and {10.0, 11.0}
#             >>> print(SingleIssueSpace(ranges=[(0.0, 1.0), (0.5, 3.0)], values={0.1, 0.9}, name='issue 3'))
#             issue 3: (0.0, 3.0)
#             >>> print(SingleIssueSpace(ranges=[(0.0, 1.0), (1.0, 2.0)], values={0.3, 1.0, 2.5}, name='issue 4'))
#             issue 4: (0.0, 2.0) and {2.5}
#         """
#         if self.everything:
#             return
#
#         self._compress_ranges()
#         self.values = set(_ for _ in self.values if not self._in_ranges(_))
#         if len(self.ranges) > 1:
#             comp = []
#             for i in range(len(self.ranges) - 1):
#                 if self.ranges[i][1] == self.ranges[i + 1][0] and self.ranges[
#                     i
#                 ][
#                     1
#                 ] in self.values:
#                     self.values.remove(self.ranges[i][1])
#                     comp.append(i)
#             for i in comp:
#                 self.ranges = self.ranges[0:i] + [
#                     (self.ranges[i][0], self.ranges[i + 1][1])
#                 ] + self.ranges[
#                     i + 2:
#                     ]
#
#     @property
#     def ancestors(self) -> List['SingleIssueSpace']:
#         """Returns all the ancestors of this issue space
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> i1 = SingleIssueSpace(count=10, name='i1')
#             >>> i1.ancestors
#             []
#             >>> i2 = SingleIssueSpace(count=10, name='i2')
#             >>> i2.ancestors
#             []
#             >>> i_1 = SingleIssueSpace(count=10, name='i_1', parent=i1)
#             >>> [_.name for _ in i_1.ancestors]
#             ['i1']
#             >>> i_12 = SingleIssueSpace(count=10, name='i_12', parent=i_1)
#             >>> [_.name for _ in i_12.ancestors]
#             ['i_1', 'i1']
#             >>> i_123 = SingleIssueSpace(count=10, name='i_123', parent=i_12)
#             >>> [_.name for _ in i_123.ancestors]
#             ['i_12', 'i_1', 'i1']
#             >>> i_2 = SingleIssueSpace(count=10, name='i_2', parent=i2)
#             >>> [_.name for _ in i_2.ancestors]
#             ['i2']
#
#
#         """
#         if self.parent is None:
#             return []
#
#         return [self.parent] + self.parent.ancestors
#
#     def __add__(self, other: Union[SingleValue, 'SingleIssueSpace']):
#         """Finds the union of two outcome spaces.
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> i1 = SingleIssueSpace(count=10, ranges= [(20.0, 30.0), (40, 50.0)], values= [50], name='i1')
#             >>> i2 = SingleIssueSpace(count=8, ranges= [(22.0, 34.0), (25, 41), (5.0, 7.0)]
#             ...                                     , values= [12, 11, 10, 13, 50], name='i2')
#             >>> print(i1 + i2)
#             (i1+i2): [(5.0, 7.0), (20.0, 50.0)] and {0, 1, 2, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 50}
#             >>> print(i1 + 144)
#             (i1+144): [(20.0, 30.0), (40, 50.0)] and {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 144, 50}
#
#         """
#         if self.everything:
#             return self.parent + other
#
#         if _is_single(other):
#             result = SingleIssueSpace(
#                 ranges=self.ranges,
#                 values=self.values,
#                 count=None,
#                 parent=self.parent,
#                 name=f'({self.name}+{other})',
#             )
#             result.values.add(other)
#         else:
#             result = SingleIssueSpace(
#                 ranges=self.ranges,
#                 values=self.values,
#                 count=None,
#                 parent=self.parent,
#                 name=f'({self.name}+{other.name})',
#             )
#             result.values = self.values.union(other.values)
#             result.ranges += other.ranges
#         result.simplify()
#         return result
#
#     def __mul__(self, other: Union[SingleValue, 'SingleIssueSpace']):
#         """Finds the intersection of two outcome spaces.
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> i1 = SingleIssueSpace(count=10, ranges= [(20.0, 30.0), (40, 50.0)], values= [50], name='i1')
#             >>> i2 = SingleIssueSpace(count=8, ranges= [(22.0, 34.0), (25, 41), (5.0, 7.0)]
#             ...                                     , values= [12, 11, 10, 13, 50], name='i2')
#             >>> print(i1 * i2)
#             (i1*i2): [(22.0, 30.0), (40, 41)] and {0, 1, 2, 3, 4, 5, 7, 50}
#             >>> print(i1 * 144)
#             (144*i1): set()
#             >>> print(i1 * 21.2)
#             (21.2*i1): {21.2}
#
#         """
#         if self.everything:
#             return self.parent * other
#
#         if _is_single(other):
#             if other in self:
#                 return SingleIssueSpace(
#                     ranges=None,
#                     values=[other],
#                     count=None,
#                     parent=self.parent,
#                     name=f'({other}*{self.name})',
#                 )
#
#             else:
#                 return SingleIssueSpace(
#                     ranges=None,
#                     values=None,
#                     count=None,
#                     parent=self.parent,
#                     name=f'({other}*{self.name})',
#                 )
#
#         result = SingleIssueSpace(
#             ranges=None,
#             values=None,
#             count=None,
#             parent=self.parent,
#             name=f'({self.name}*{other.name})',
#         )
#         result.values = self.values.intersection(other.values)
#         for r1 in self.ranges:
#             for r2 in other.ranges:
#                 intersection = _intersect_tuples(r1, r2)
#                 result.ranges.append(intersection)
#         result.simplify()
#         return result
#
#     def __sub__(self, other: Union[SingleValue, 'SingleIssueSpace']):
#         """An outcome space containing this outcome space minus the intersection with other.
#
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> i1 = SingleIssueSpace(count=10, ranges= [(20.0, 30.0), (40, 50.0)], values= [50], name='i1')
#             >>> i2 = SingleIssueSpace(count=8, ranges= [(22.0, 34.0), (25, 41), (5.0, 7.0)]
#             ...                                     , values= [12, 11, 10, 13, 50], name='i2')
#             >>> print(i1 - i2)
#             (i1-i2): [(20.0, 22.0), (41, 50.0)] and {8, 9, 6}
#             >>> print(i1 - 144)
#             (i1-144): [(20.0, 30.0), (40, 50.0)] and {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 50}
#             >>> print(i1 - 21.2)
#             (i1-21.2): [(20.0, 21.2), (21.2, 30.0), (40, 50.0)] and {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 50}
#
#
#
#         """
#         if self.everything:
#             return self.parent - other
#
#         if _is_single(other):
#             ranges = []
#             for r in self.ranges:
#                 if r[0] < other < r[1]:
#                     ranges.append((r[0], other))
#                     ranges.append((other, r[1]))
#                 else:
#                     ranges.append(r)
#             values = self.values.copy()
#             if other in values:
#                 values.remove(other)
#             return SingleIssueSpace(
#                 ranges=ranges,
#                 values=values,
#                 count=None,
#                 parent=self.parent,
#                 name=f'({self.name}-{other})',
#             )
#
#         result = SingleIssueSpace(
#             ranges=None,
#             values=None,
#             count=None,
#             parent=self.parent,
#             name=f'({self.name}-{other.name})',
#         )
#         result.values = self.values.difference(other.values)
#         other *= self
#         for r1 in self.ranges:
#             survived = r1
#             for r2 in other.ranges:
#                 if isinstance(survived, tuple):
#                     survived = _subtract_tuples(survived, r2)
#                 else:
#                     survived = [_subtract_tuples(_, r2) for _ in survived]
#             if isinstance(survived, tuple):
#                 survived = [survived]
#             result.ranges += survived
#         result.simplify()
#         return result
#
#     @property
#     def empty(self):
#         """Checks if this is an empty SingleIssueSpace.
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> SingleIssueSpace().empty
#             True
#             >>> SingleIssueSpace(everything=True).empty
#             True
#             >>> i1 = SingleIssueSpace(count=10)
#             >>> i1.empty
#             False
#             >>> SingleIssueSpace(everything=True, parent=i1).empty
#             False
#
#         """
#         if self.everything:
#             if self.parent is None:
#                 return True
#
#             return self.parent.empty
#
#         return len(self.ranges) == 0 and len(self.values) == 0
#
#     def __neg__(self) -> 'SingleIssueSpace':
#         """The inverse of this outcome space. Assumes that a parent outcome space is defined.
#
#         Examples:
#             >>> from negmas import SingleIssueSpace
#             >>> i1 = SingleIssueSpace(count=10, ranges= [(20.0, 30.0), (40, 50.0)], values= [50], name='i1')
#             >>> i2 = SingleIssueSpace(count=8, ranges= [(22.0, 34.0), (25, 41), (5.0, 7.0)]
#             ...                                     , values= [12, 11, 10, 13, 50], name='i2')
#             >>> try:
#             ...     print(-i2)
#             ... except RuntimeError:
#             ...     print('cannot inverse without a parent')
#             cannot inverse without a parent
#             >>> i2 = SingleIssueSpace(count=8, ranges= [(22.0, 34.0), (25, 41), (5.0, 7.0)]
#             ...                                     , values= [12, 11, 10, 13, 50], parent=i1, name='i2')
#             >>> print(-i2)
#             -i2: [(20.0, 22.0), (41, 50.0)] and {8, 9, 6}
#
#
#         """
#         if self.everything:
#             return self.parent if self.parent is not None else SingleIssueSpace(
#                 ranges=None,
#                 values=None,
#                 count=None,
#                 parent=None,
#                 name=f'-{self.name}',
#             )
#
#         if self.empty:
#             return self.parent if self.parent is not None else SingleIssueSpace(
#                 ranges=None,
#                 values=None,
#                 count=None,
#                 parent=None,
#                 everything=True,
#                 name=f'-{self.name}',
#             )
#
#         if self.parent is None:
#             raise RuntimeError(
#                 'Cannot find the inverse of an issue space that has no parent'
#             )
#
#         result = self.parent - self
#         result.rename(f'-{self.name}')
#         return result
#
#     def rand(self) -> Optional[SingleValue]:
#         """Picks a random valid value.
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> i1 = SingleIssueSpace(count=10, ranges= [(20.0, 30.0), (40, 50.0)], values= [50], name='i1')
#             >>> samples = [i1.rand() for _ in range(10)]
#             >>> [_ in i1 for _ in samples]
#             [True, True, True, True, True, True, True, True, True, True]
#
#         """
#         if self.everything:
#             if self.parent is None:
#                 return None
#
#             return self.parent.rand()
#
#         if self.empty:
#             return None
#
#         if len(self.ranges) == 0:
#             return random.sample(self.values)
#
#         sizes = [_[1] - _[0] for _ in self.ranges]
#         total_size = sum(sizes)
#         v = random.random() * total_size
#         total = 0.0
#         for s, r in zip(sizes, self.ranges):
#             total += s
#             if v > total:
#                 return r[0] + v - total
#
#         return self.ranges[-1][1] - EPSILON
#
#     rand_valid = rand
#
#     def rand_invalid(self) -> Optional[SingleValue]:
#         """Pick a random *invalid* value
#
#         Examples:
#
#             >>> from negmas import SingleIssueSpace
#             >>> i1 = SingleIssueSpace(count=10, ranges= [(20.0, 30.0), (40, 50.0)], values= [50], name='i1')
#             >>> samples = [i1.rand_invalid() for _ in range(10)]
#             >>> [_ in i1 for _ in samples]
#             [False, False, False, False, False, False, False, False, False, False]
#
#
#         """
#         if self.everything:
#             if self.parent is None:
#                 return None
#
#             return self.parent.rand_invalid()
#
#         if self.empty:
#             return 'anything'
#
#         if len(self.ranges) == 0:
#             if isinstance(random.sample(self.values), str):
#                 return unique_name('', add_time=False)
#
#             mx = max(self.values)
#             return random.randint(mx + 1, 2 * mx)
#
#         mx = self.ranges[0][1]
#         mn = self.ranges[0][0]
#         for r in self.ranges:
#             mx = max([mx, r[1]])
#             mn = min([mn, r[1]])
#         return random.random() * (mx - mn) + mx
#
#     def __str__(self):
#         """Returns a string representation.
#
#         Examples:
#
#             >>> i1 = SingleIssueSpace(ranges=(0.0, 1.0), values=[2.0, 3.0, 4.0], name='issue 1')
#             >>> print(i1)
#             issue 1: (0.0, 1.0) and {2.0, 3.0, 4.0}
#             >>> print(SingleIssueSpace(ranges=(0.0, 1.0), name='issue 2'))
#             issue 2: (0.0, 1.0)
#             >>> print(SingleIssueSpace(ranges=(0.0, 1.0), name='issue 3', parent=i1))
#             issue 3 (parent: issue 1): (0.0, 1.0)
#             >>> print(SingleIssueSpace(parent=i1, name='issue 4'))
#             issue 4 (parent: issue 1): set()
#             >>> print(SingleIssueSpace(parent=i1, everything=True, name='issue 4'))
#             issue 4 (parent: issue 1): everything
#             >>> print(SingleIssueSpace(values={0.0, 1.0}, name='issue 5'))
#             issue 5: {0.0, 1.0}
#             >>> print(SingleIssueSpace(ranges=[(0.0, 1.0), (2.0, 3.0)], values={10.0, 11.0}, name='issue 6'))
#             issue 6: [(0.0, 1.0), (2.0, 3.0)] and {10.0, 11.0}
#             >>> print(SingleIssueSpace(ranges=[(0.0, 1.0), (0.5, 3.0)], values={10.0, 11.0}, name='issue 6'))
#             issue 6: (0.0, 3.0) and {10.0, 11.0}
#             >>> print(SingleIssueSpace(ranges=[(0.0, 1.0), (0.5, 3.0)], values={0.1, 0.9}, name='issue 6'))
#             issue 6: (0.0, 3.0)
#
#         """
#         if self.everything:
#             base = f'everything'
#         elif len(self.ranges) == 0:
#             base = f'{self.values}'
#         elif len(self.values) == 0:
#             if len(self.ranges) == 1:
#                 base = f'{self.ranges[0]}'
#             else:
#                 base = f'{self.ranges}'
#         else:
#             if len(self.ranges) == 1:
#                 base = f'{self.ranges[0]} and {self.values}'
#             else:
#                 base = f'{self.ranges} and {self.values}'
#         if self.parent is not None:
#             parent = f' (parent: {self.parent.name})'
#         else:
#             parent = ''
#         return f'{self.name}{parent}: {base}'
#
#     def __and__(self, other):
#         return self * other
#
#     def __or__(self, other):
#         return self + other
#
#     def __xor__(self, other):
#         return self + other - (self * other)
#
#     def __invert__(self):
#         return -self
#
#     def __mod__(self, other):
#         return self - other
#
#     cardinality = n_outcomes
#     dimensionality = 1
#     n_dims = 1
#     contains = __contains__
#     union = __add__
#     instersection = __mul__
#     difference = __sub__

# class OutcomeSpace(object):
#     """An arbitrary constraint on outcomes.
#
#     Overrides standard operators so it can be used something like the following:
#
#
#
#     """
#     __slots__ = ('_parent', '_range')
#     def __init__(self, parent: Optional['OutcomeSpace']=None, **kwargs) -> None:
#         super().__init__()
#         self._parent = parent
#         self._range = copy.deepcopy(kwargs)
#
#     @property
#     def n_augmented_outcomes(self) -> Union[int, float]:
#         """Returns the total number of outcomes in a set of issues. `-1` indicates infinity"""
#
#         n = 1
#         for dim in self._range.values():
#             n *= _dim_cardinality(dim)
#             if math.isinf(n):
#                 return math.inf
#         return n
#
#     cardinality = n_augmented_outcomes
#
#     @property
#     def n_dims(self):
#         return len(self._range)
#
#     dimensionality = n_dims
#
#     @property
#     def cardinalities(self) -> Dict[str, Union[int, float]]:
#         """The cardinality of all dimensions"""
#         return {k: _dim_cardinality(cutoff_utility) for k, cutoff_utility in self._range}
#
#     @property
#     def types(self) -> Dict[str, str]:
#         """The type of the issue.
#
#         Returns:
#             str: either 'continuous' or 'discrete'
#
#         """
#         return {k: 'continuous' if math.isinf(_dim_cardinality(cutoff_utility)) else 'discrete' for k, cutoff_utility in self._range}
#
#     def is_infinite(self) -> bool:
#         """Test whether any issue is continuous (infinite outcome space)
#
#         Returns:
#             bool: continuous or not
#
#         """
#         return any(math.isinf(_) for _ in self.cardinalities)
#
#     def is_finite(self) -> bool:
#         """Test whether all issues are discrete (finite outcome space)
#
#         Returns:
#             bool: discrete or not
#
#         """
#         return not self.is_infinite()
#
#     @property
#     def all(self) -> Generator:
#         """A generator that generates all possible values.
#
#         Remarks:
#             - This function returns a generator for the case when the number of values is very large.
#             - If you need a list then use something like:
#
#
#         """
#         if self.is_infinite():
#             raise ValueError('Cannot return all possibilities of a continuous issues')
#         if len(self._range) == 0:
#             raise StopIteration
#         yield from itertools.product(_all_values for _ in self._range)
#
#     def __contains__(self, item: Union[Outcome, OutcomeRange, 'OutcomeSpace']):
#         """Tests if an outcome, outcome range, or another outcome space is within this outcome space."""
#         return outcome_in_range(item, self._range)
#
#     def __add__(self, other: Union[Outcome, OutcomeRange, 'OutcomeSpace']):
#         """Finds the union of two outcome spaces."""
#         r = OutcomeSpace(parent=self._parent)
#         r._range = range_union(self._range, other._range)
#         return r
#
#     def __mul__(self, other: Union[Outcome, OutcomeRange, 'OutcomeSpace']):
#         """Finds the intersection of two outcome spaces."""
#         r = OutcomeSpace(parent=self._parent)
#         r._range = range_intersection(self._range, other._range)
#         return r
#
#     def __sub__(self, other: Union[Outcome, OutcomeRange, 'OutcomeSpace']):
#         """An outcome space containing this outcome space minus the intersection with other."""
#
#     def __neg__(self):
#         """The inverse of this outcome space. Assumes that a parent outcome space is defined."""
#
#     def rand(self) -> Dict[str, Union[int, float, str]]:
#         """Picks a random valid value."""
#
#     rand_valid = rand
#
#     def rand_invalid(self) -> Dict[str, Union[int, float, str]]:
#         """Pick a random *invalid* value"""


def outcome_as_dict(outcome: Outcome, issue_names: List[str] = None):
    """Converts the outcome to a dict no matter what was its type"""
    if outcome is None:
        return None
    if isinstance(outcome, dict):
        return outcome
    if isinstance(outcome, OutcomeType):
        return outcome.asdict()
    if issue_names is not None:
        return dict(zip(issue_names, outcome))
    return dict(zip((str(_) for _ in range(len(outcome))), outcome))


def outcome_as_tuple(outcome: Outcome):
    """Converts the outcome to a tuple no matter what was its type"""
    if outcome is None:
        return None
    if isinstance(outcome, tuple):
        return outcome
    if isinstance(outcome, OutcomeType):
        return outcome.astuple()
    if isinstance(outcome, dict):
        return list(outcome.values())
    raise ValueError(f'Unknown type for outcome {type(outcome)}')



import pathlib

import numpy as np
import pkg_resources
import pytest

from negmas import GeniusNegotiator, load_genius_domain, load_genius_domain_from_folder, genius_bridge_is_running

dom_folder = pathlib.Path(pkg_resources.resource_filename('negmas'
                                                          , resource_name='tests/data/scenarios/anac/y2010/Travel'))
dom = dom_folder / 'travel_domain.xml'
util1 = dom_folder / 'travel_chox.xml'
util2 = dom_folder / 'travel_fanny.xml'


@pytest.fixture(scope='module')
def init_genius():
    pass


@pytest.mark.skipif(condition=not genius_bridge_is_running(), reason='No Genius Bridge, skipping genius-agent tests')
def test_genius_agent(init_genius):
    p, _, issues = load_genius_domain_from_folder(dom_folder
                                                  , agent_factories=[
            lambda: GeniusNegotiator(java_class_name='agents.anac.y2015.Atlas3.Atlas3'
                                     , domain_file_name=dom, utility_file_name=util1),
            lambda: GeniusNegotiator(java_class_name='agents.anac.y2015.AgentX.AgentX'
                                     , domain_file_name=dom, utility_file_name=util2)]
                                                  , keep_issue_names=True, keep_value_names=True, time_limit=20)
    assert len(p.negotiators) > 1
    issue_list = [f'{k}:{v}' for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals'"
        ", 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments'"
        ", 'Special streets', 'Palace', 'Landscape and nature']"]
    atlas3, agentx = p.negotiators[0], p.negotiators[1]
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(atlas3.utility_function(s.current_offer)) for s in p.history])
    u2 = np.array([float(agentx.utility_function(s.current_offer)) for s in p.history])
    welfare = u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)
    assert len(u1) > 0


@pytest.mark.skipif(condition=not genius_bridge_is_running(), reason='No Genius Bridge, skipping genius-agent tests')
def test_genius_agent_step_limit(init_genius):
    p, _, issues = load_genius_domain_from_folder(dom_folder
                                                  , agent_factories=[
            lambda: GeniusNegotiator(java_class_name='agents.anac.y2015.Atlas3.Atlas3'
                                     , domain_file_name=dom, utility_file_name=util1),
            lambda: GeniusNegotiator(java_class_name='agents.anac.y2015.AgentX.AgentX'
                                     , domain_file_name=dom, utility_file_name=util2)]
                                                  , keep_issue_names=True, keep_value_names=True, n_steps=20,
                                                  time_limit=None)
    issue_list = [f'{k}:{v}' for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)', 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals', 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments', 'Special streets', 'Palace', 'Landscape and nature']"]
    atlas3, agentx = p.negotiators[0], p.negotiators[1]
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(atlas3.utility_function(s.current_offer)) for s in p.history])
    u2 = np.array([float(agentx.utility_function(s.current_offer)) for s in p.history])
    welfare = u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)


@pytest.mark.skipif(condition=not genius_bridge_is_running(), reason='No Genius Bridge, skipping genius-agent tests')
def test_genius_agent_step_long_session(init_genius):
    a1 = GeniusNegotiator(java_class_name="agents.anac.y2015.Atlas3.Atlas3"
                          , domain_file_name=dom, utility_file_name=util1)
    a2 = GeniusNegotiator(java_class_name="agents.anac.y2015.Atlas3.Atlas3"
                          , domain_file_name=dom, utility_file_name=util2)
    p, _, issues = load_genius_domain(dom, keep_issue_names=True
                                      , keep_value_names=True
                                      , n_steps=20, time_limit=None)
    issue_list = [f'{k}:{v}' for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)', 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo', 'Festivals', 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.', 'Monuments', 'Special streets', 'Palace', 'Landscape and nature']"]
    p.add(a1)
    p.add(a2)
    p.run()
    # print(f'{len(p.history)} bids exchanged')
    u1 = np.array([float(a1.utility_function(s.current_offer)) for s in p.history])
    u2 = np.array([float(a2.utility_function(s.current_offer)) for s in p.history])
    welfare = u1 + u2
    # print(
    #     f'Negotiator 1: {u1.mean()}({u1.std()})[{u1.min()}, {u1.max()}]\nNegotiator 2: {u2.mean()}({u2.std()})[{u1.min()}, {u1.max()}]'
    #     f'\nWelfare: {welfare.mean()}({welfare.std()})[{welfare.min()}, {welfare.max()}]')
    # print(p.state)
    assert len(u1) > 4


@pytest.mark.skipif(condition=not genius_bridge_is_running(), reason='No Genius Bridge, skipping genius-agent tests')
def test_genius_agent_same_utility(init_genius):
    from negmas import GeniusNegotiator, load_genius_domain
    dom = dom_folder / 'travel_domain.xml'
    util1 = dom_folder / 'travel_chox.xml'
    util2 = util1
    a1 = GeniusNegotiator(java_class_name="agents.anac.y2015.Atlas3.Atlas3"
                          , domain_file_name=dom, utility_file_name=util1)
    a2 = GeniusNegotiator(java_class_name="agents.anac.y2015.AgentX.AgentX"
                          , domain_file_name=dom, utility_file_name=util2)
    p, _, issues = load_genius_domain(dom, keep_issue_names=True
                                      , keep_value_names=True
                                      , time_limit=30)
    issue_list = [f'{k}:{v}' for k, v in enumerate(issues)]
    assert issue_list == [
        "0:Atmosphere: ['Cultural heritage', 'Local traditions', 'Political stability', 'Security (personal)'"
        ", 'Liveliness', 'Turistic activities', 'Hospitality']",
        "1:Amusement: ['Nightlife and entertainment', 'Nightclubs', 'Excursion', 'Casinos', 'Zoo'"
        ", 'Festivals', 'Amusement park']",
        "2:Culinary: ['Local cuisine', 'Lunch facilities', 'International cuisine', 'Cooking workshops']",
        "3:Shopping: ['Shopping malls', 'Markets', 'Streets', 'Small boutiques']",
        "4:Culture: ['Museum', 'Music hall', 'Theater', 'Art gallery', 'Cinema', 'Congress center']",
        "5:Sport: ['Bike tours', 'Hiking', 'Indoor activities', 'Outdoor activities', 'Adventure']",
        "6:Environment: ['Parks and Gardens', 'Square', 'Historical places', 'See, river, etc.'"
        ", 'Monuments', 'Special streets', 'Palace', 'Landscape and nature']"]
    p.add(a1)
    p.add(a2)
    p.run()
    u1 = np.array([float(a1.utility_function(s.current_offer)) for s in p.history])
    u2 = np.array([float(a2.utility_function(s.current_offer)) for s in p.history])
    welfare = u1 + u2
    assert len(u1) == 2
    assert welfare[0] == welfare[1] == 2.0
    assert p.state.agreement is not None
    assert p.state.broken is False


class TestGeniusAgentSessions:
    def prepare(self, utils=(1, 1), single_issue=True, keep_issue_names=True,
                keep_value_names=True):
        from negmas import convert_genius_domain_from_folder
        src = pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        dst = pkg_resources.resource_filename('negmas', resource_name='tests/data/LaptopConv1D')
        if single_issue:
            assert convert_genius_domain_from_folder(src_folder_name=src
                                                     , dst_folder_name=dst
                                                     , force_single_issue=True
                                                     , cache_and_discretize_outcomes=True
                                                     , n_discretization=10
                                                     )
            base_folder = dst
        else:
            base_folder = src
        neg, agent_info, issues = load_genius_domain_from_folder(base_folder
                                                                 , keep_issue_names=keep_issue_names
                                                                 , keep_value_names=keep_value_names)
        # atlas = GeniusNegotiator.random_negotiator(
        atlas = GeniusNegotiator(java_class_name='agents.anac.y2015.Atlas3.Atlas3',
                                 domain_file_name=base_folder + '/Laptop-C-domain.xml'
                                 , utility_file_name=base_folder + f'/Laptop-C-prof{utils[0]}.xml'
                                 , keep_issue_names=keep_issue_names
                                 , keep_value_names=keep_value_names)
        # agentx = GeniusNegotiator.random_negotiator(
        agentx = GeniusNegotiator(java_class_name='agents.anac.y2015.AgentX.AgentX',
                                  domain_file_name=base_folder + '/Laptop-C-domain.xml'
                                  , utility_file_name=base_folder + f'/Laptop-C-prof{utils[1]}.xml'
                                  , keep_issue_names=keep_issue_names
                                  , keep_value_names=keep_value_names
                                  )
        neg.add(atlas)
        neg.add(agentx)
        return neg

    @pytest.mark.skipif(condition=not genius_bridge_is_running(),
                        reason='No Genius Bridge, skipping genius-agent tests')
    def test_genius_agents_can_run_on_converted_single_issue_ufun1(self, init_genius):
        neg = self.prepare(utils=(1, 1), single_issue=True)
        assert neg.pareto_frontier(sort_by_welfare=True)[0] == [(1.0, 1.0)]
        state = neg.run()
        # pprint(neg.history)
        assert neg.agreement is not None
        #        assert len(neg.history) <= 3
        assert neg.agreement == {'Laptop-Harddisk-External Monitor': "HP+60 Gb+19'' LCD"}

    @pytest.mark.skipif(condition=not genius_bridge_is_running(),
                        reason='No Genius Bridge, skipping genius-agent tests')
    def test_genius_agents_can_run_on_converted_single_issue_ufun2(self, init_genius):
        neg = self.prepare(utils=(2, 2), single_issue=True)
        assert neg.pareto_frontier(sort_by_welfare=True)[0] == [(1.0, 1.0)]
        state = neg.run()
        assert neg.agreement is not None
        assert len(neg.history) <= 3
        assert neg.agreement == {'Laptop-Harddisk-External Monitor': "Macintosh+80 Gb+19'' LCD"}

    @pytest.mark.skipif(condition=not genius_bridge_is_running(),
                        reason='No Genius Bridge, skipping genius-agent tests')
    def test_genius_agents_can_run_on_converted_single_issue(self, init_genius):
        neg = self.prepare(utils=(1, 2), single_issue=True)
        assert neg.pareto_frontier(sort_by_welfare=True)[0] == [(0.7715533992081258, 0.8450562871935449),
                                                                (0.5775524426410947, 1.0),
                                                                (1.0, 0.5136317604069089),
                                                                (0.8059990434329689, 0.6685754732133642)
                                                                ]

        state = neg.run()
        # assert len(neg.history) >= 2
        assert neg.agreement is not None

    @pytest.mark.skipif(condition=not genius_bridge_is_running(),
                        reason='No Genius Bridge, skipping genius-agent tests')
    def test_genius_agents_can_run_on_converted_multiple_issues(self, init_genius):
        neg = self.prepare(utils=(1, 1), single_issue=False)
        frontier = neg.pareto_frontier(sort_by_welfare=True)[0]
        true_frontier = [(1.0, 1.0)]
        assert len(frontier) == len(true_frontier)
        for a, b in zip(frontier, true_frontier):
            assert abs(a[0] - b[0]) < 1e-5 and abs(a[1] - b[1]) < 1e-5

        state = neg.run()
        assert len(neg.history) < 3
        assert neg.agreement is not None
        assert neg.agreement == {'Laptop': 'HP', 'Harddisk': '60 Gb', 'External Monitor': "19'' LCD"}

    @pytest.mark.skipif(condition=not genius_bridge_is_running(),
                        reason='No Genius Bridge, skipping genius-agent tests')
    def test_genius_agents_can_run_on_converted_multiple_issues_no_names(self, init_genius):
        neg = self.prepare(utils=(1, 1), single_issue=False, keep_issue_names=False)
        frontier = neg.pareto_frontier(sort_by_welfare=True)[0]
        true_frontier = [(1.0, 1.0)]
        assert len(frontier) == len(true_frontier)
        for a, b in zip(frontier, true_frontier):
            assert abs(a[0] - b[0]) < 1e-5 and abs(a[1] - b[1]) < 1e-5

        state = neg.run()
        assert len(neg.history) < 3
        assert neg.agreement is not None
        assert neg.agreement == ('HP', '60 Gb', "19'' LCD")

    @pytest.mark.skipif(condition=not genius_bridge_is_running(),
                        reason='No Genius Bridge, skipping genius-agent tests')
    def test_genius_agent_example(self, init_genius):
        agent_name1 = 'agents.anac.y2015.Atlas3.Atlas3'
        agent_name2 = 'agents.anac.y2015.Atlas3.Atlas3'
        single_issue = False
        keep_issue_names, keep_value_names = False, False
        utils = (1, 1)

        from negmas import convert_genius_domain_from_folder
        src = pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        dst = pkg_resources.resource_filename('negmas', resource_name='tests/data/LaptopConv1D')
        if single_issue:
            assert convert_genius_domain_from_folder(src_folder_name=src
                                                     , dst_folder_name=dst
                                                     , force_single_issue=True
                                                     , cache_and_discretize_outcomes=True
                                                     , n_discretization=10
                                                     )
            base_folder = dst
        else:
            base_folder = src
        neg, agent_info, issues = load_genius_domain_from_folder(base_folder
                                                                 , keep_issue_names=keep_issue_names
                                                                 , keep_value_names=keep_value_names)
        # atlas = GeniusNegotiator.random_negotiator(
        atlas = GeniusNegotiator(java_class_name=agent_name1,
                                 domain_file_name=base_folder + '/Laptop-C-domain.xml'
                                 , utility_file_name=base_folder + f'/Laptop-C-prof{utils[0]}.xml'
                                 , keep_issue_names=keep_issue_names
                                 , keep_value_names=keep_value_names)
        agentx = GeniusNegotiator(java_class_name=agent_name2,
                                  domain_file_name=base_folder + '/Laptop-C-domain.xml'
                                  , utility_file_name=base_folder + f'/Laptop-C-prof{utils[1]}.xml'
                                  , keep_issue_names=keep_issue_names
                                  , keep_value_names=keep_value_names
                                  )
        neg.add(atlas)
        neg.add(agentx)
        # print(agent_name1, agent_name2, neg.run(timeout=1))


if __name__ == '__main__':
    pytest.main(args=[__file__])



#!/usr/bin/env python
"""A set of utilities that can be used by agents developed for the platform.

This set of utlities can be extended but must be backward compatible for at
least two versions
"""
import datetime
import importlib
import json
import logging
import math
import pathlib
import string
import sys
from typing import List, Optional, Iterable, Union, Callable, Mapping, Any, Sequence, Tuple, Dict, Type
from typing import TYPE_CHECKING

from negmas import NEGMAS_CONFIG

if TYPE_CHECKING:
    pass

import colorlog
import numpy as np
import os
import random
import re
import scipy.stats as stats
import inflect
import stringcase
from enum import Enum
import yaml
from negmas.generics import *
import copy

__all__ = [
    'create_loggers',
    # 'MultiIssueUtilityFunctionMapping',
    'ReturnCause',
    'Distribution',  # A probability distribution
    'snake_case',
    'camel_case',
    'unique_name',
    'is_nonzero_file',
    'pretty_string',
    'ConfigReader',
    'get_class',
    'import_by_name',
    'get_full_type_name',
    'instantiate',
    'humanize_time',
    'gmap', 'ikeys', 'Floats',
    'DEFAULT_DUMP_EXTENSION', 'dump'
]
# conveniently named classes

"""Maps from a single issue to a Negotiator function."""
# MultiIssueUtilityFunctionMapping = Union[
#    Callable[['Issues'], 'UtilityFunction'], Mapping['Issues', 'UtilityFunction']]  # type: ignore
# """Maps between multiple issues and a Negotiator function."""
ParamList = List[Union[int, str]]
GenericMappings = List[GenericMapping]
IterableMappings = List[IterableMapping]
# MultiIssueUtilityFunctionMappings = List[MultiIssueUtilityFunctionMapping]
ParamLists = Iterable[ParamList]
Floats = List[float]

COMMON_LOG_FILE_NAME = './logs/{}_{}.txt'.format('log', datetime.datetime.now().strftime('%Y%m%d-%H%M%S'))

MODULE_LOG_FILE_NAME: Dict[str, str] = dict()

LOGS_BASE_DIR = './logs'

DEFAULT_DUMP_EXTENSION = NEGMAS_CONFIG.get('default_dump_extension', 'json')


class ReturnCause(Enum):
    TIMEOUT = 0
    SUCCESS = 1
    FAILURE = 2


def create_loggers(
    file_name: Optional[str] = None,
    module_name: Optional[str] = None,
    screen_level: Optional[int] = logging.WARNING,
    file_level: Optional[int] = logging.DEBUG,
    format_str: str = '%(asctime)s - %(levelname)s - %(message)s',
    colored: bool = True,
    app_wide_log_file: bool = True,
    module_wide_log_file: bool = False,
) -> logging.Logger:
    """
    Create a set of loggers to report feedback.

    The logger created can log to both a file and the screen at the  same time
    with adjustable level for each of them. The default is to log everything to
    the file and to log WARNING at least to the screen

    Args:
        module_wide_log_file:
        app_wide_log_file:
        file_name: The file to export_to the logs to. If None only the screen
                    is used for logging. If empty, a time-stamp is used
        module_name: The module name to use. If not given the file name
                    without .py is used
        screen_level: level of the screen logger
        file_level: level of the file logger
        format_str: the format of logged items
        colored: whether or not to try using colored logs

    Returns:
        logging.Logger: The logger

    """
    if module_name is None:
        module_name = __file__.split('/')[-1][:-3]
    # create logger if it does not already exist
    logger = logging.getLogger(module_name)
    if len(logger.handlers) > 0:
        return logger
    logger.setLevel(logging.DEBUG)
    # create formatter
    if colored and 'colorlog' in sys.modules and os.isatty(2):
        date_format = '%Y-%m-%d %H:%M:%S'
        cformat = '%(log_color)s' + format_str
        formatter = colorlog.ColoredFormatter(
            cformat,
            date_format,
            log_colors={
                'DEBUG': 'reset',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'bold_red',
            },
        )
    else:
        formatter = logging.Formatter(format_str)
    if screen_level is not None:
        # create console handler and set level to logdebug
        screen_logger = logging.StreamHandler()
        screen_logger.setLevel(screen_level)
        # add formatter to ch
        screen_logger.setFormatter(formatter)
        # add ch to logger
        logger.addHandler(screen_logger)
    if file_name is not None and file_level is not None:
        file_name = str(file_name)
        if len(file_name) == 0:
            if app_wide_log_file:
                file_name = COMMON_LOG_FILE_NAME
            elif module_wide_log_file and module_name in MODULE_LOG_FILE_NAME.keys():
                file_name = MODULE_LOG_FILE_NAME[module_name]
            else:
                file_name = '{}/{}_{}.txt'.format(LOGS_BASE_DIR,
                    module_name, datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
                )
                MODULE_LOG_FILE_NAME[module_name] = file_name

            os.makedirs(f'{LOGS_BASE_DIR}', exist_ok=True)
        os.makedirs(os.path.dirname(file_name), exist_ok=True) # type: ignore
        file_logger = logging.FileHandler(file_name)
        file_logger.setLevel(file_level)
        file_logger.setFormatter(formatter)
        logger.addHandler(file_logger)
    return logger


def snake_case(s: str) -> str:
    """ Converts a string from CamelCase to snake_case

    Example:

        >>> print(snake_case('ThisIsATest'))
        this_is_a_test



    Args:
        s: input string

    Returns:
        str: converted string
    """
    return re.sub(
        '(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', s
    ).lower().strip(
        '_'
    )


def camel_case(s: str, capitalize_first: bool = False, lower_first: bool = False) -> str:
    """ Converts a string from snake_case to CamelCase

    Example:

        >>> print(camel_case('this_is_a_test'))
        thisIsATest
        >>> print(camel_case('this_is_a_test', capitalize_first=True))
        ThisIsATest
        >>> print(camel_case('This_is_a_test', lower_first=True))
        thisIsATest
        >>> print(camel_case('This_is_a_test'))
        ThisIsATest

    Args:
        s: input string
        capitalize_first: if true, the first character will be capitalized
        lower_first: If true, the first character will be lowered

    Returns:
        str: converted string
    """
    if len(s) < 1:
        return s
    parts = s.split('_')
    if capitalize_first:
        parts = [_.capitalize() for _ in parts]
    elif lower_first:
        parts = [parts[0].lower()] +  [_.capitalize() for _ in parts[1:]]
    else:
        parts = [parts[0]] + [_.capitalize() for _ in parts[1:]]

    return ''.join(parts)


def unique_name(base: str, add_time=True, rand_digits=8) -> str:
    """Return a unique name.

    Can be used to return a unique directory name on the givn base.

    Args:
        base: str (str): base path/string
        add_time (bool, optional): Defaults to True. Add current time
        rand_digits (int, optional): Defaults to 8. The number of random
            characters to add to the name

    Examples:

        >>> a = unique_name('')
        >>> len(a) == 8 + 1 + 6 + 8
        True

    Returns:
        str: The unique name.

    """
    if rand_digits > 0:
        characters = string.ascii_letters + string.digits
        password = "".join(
            random.choice(characters) for _ in range(rand_digits)
        )
    else:
        password = ''
    if add_time:
        _time = datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
    else:
        _time = ''
    sub = _time + password
    return os.path.join(base, sub) if len(sub) > 0 else base


def is_nonzero_file(fpath: str) -> bool:
    """Whether or not the path is for an existing nonzero file.

    Args:
        fpath: path to the file to test. It accepts both str and pathlib.Path

    """
    return os.path.isfile(fpath) and os.path.getsize(fpath) > 0


def pretty_string(src: Any, tab_size=2, compact=False) -> str:
    """Recursively print nested elements.

        Args:
            src (Any): The source to be converted to a printable string
            tab_size (int): Tab size in spaces
            compact (bool): If true the output is  converted into a single line

        Returns:
            str: The pretty version of the input

        Remarks:
            - This function assumes that the patterns `` "`` and ``":`` do not appear anywhere in the input.
              If they appear, the space, : will be removed.
        """
    s = _pretty_string(src, dpth=0, current_key='', tab_size=tab_size)
    if compact:
        return s.replace('\n', '')

    else:
        return s.replace(' "', ' ').replace('":', ':')


def _pretty_string(src, dpth=0, current_key='', tab_size=2) -> str:
    """Recursively print nested elements.

    Args:
        dpth (int): Current depth
        current_key (str): Current key being printed
        tab_size: Tab size in spaces

    Returns:
        str: The pretty version of the input
    """

    def tabs(n):
        return ' ' * n * tab_size  # or 2 or 8 or...

    output = ''
    if isinstance(src, dict):
        output += tabs(dpth) + '{\n'
        for key, value in src.items():
            output += _pretty_string(value, dpth + 1, key) + '\n'
        output += tabs(dpth) + '}'
    elif isinstance(src, list) or isinstance(src, tuple):
        output += tabs(dpth) + '[\n'
        for litem in src:
            output += _pretty_string(litem, dpth + 1) + '\n'
        output += tabs(dpth) + ']'
    else:
        if len(current_key) > 0:
            output += (tabs(dpth) + '"%s":%s' % (current_key, src))
        else:
            output += (tabs(dpth) + '%s' % src)
    return output


class LazyInitializable(object):
    """Base Negotiator for all agents

    Supports a set_params function that can be used for lazy initialization
    """

    def __init__(self) -> None:
        super().__init__()

    def set_params(self, **kwargs) -> None:
        """Sets the attributes of the object.

        This function can be used to set the attributes of any object to the
        same values used in its construction which allows for lazy
        initialization.

        Args:
            **kwargs: The parameters usually passed to the constructor as a dict

        Example:

            >>> class A(LazyInitializable):
            ...     def __init__(self, a=None, b=None) -> None:
            ...         super().__init__()
            ...         self.a = a
            ...         self.b = b

            Now you can do the following::

            >>> a = A()
            >>> a.set_params(a=3, b=2)

            which will be equivalent to:

            >>> b = A(a=3, b=2)

        Remarks:
            - See ``adjust_params()`` for an example in which the constuctor needs to do more processing than just
              assinging its inputs to instance members.

        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.adjust_params()

    def adjust_params(self) -> None:
        """Adjust the internal attributes following ``set_attributes()`` or construction using ``__init__()``.

        This function needs to be implemented only if the constructor needs to
        do some processing on the inputs other than assigning it to instance
        attributes. In such case, move these adjustments to this function and
        call it in the constructor.

        Examples:

            >>> class A(object):
            ...     def __init__(self, a=None, b=None):
            ...         self.a = a
            ...         self.b = b if b is not None else []

            should now be defined as follows:

            >>> class A(LazyInitializable):
            ...     def __init__(self, a, b):
            ...         super().__init__()
            ...         self.a = a
            ...         self.b = b
            ...         self.adjust_params()
            ...
            ...     def adjust_params(self):
            ...         if self.b is None: self.b = []

        Remarks:
            - Remember to call `super().__init__()` first in your constructor and to call your `adjust_params()` by
              the end of the constructor.
            - The constructor should ONLY copy the parameters it receives to internal variables and then calls
              `adjust_params()` if any more processing is needed. This makes it possible to use `set_params()` with
              this object.
            - You should **never** call `adjust_params()` directly anywhere.
        """
        pass


class Distribution(object):
    """Any distribution from scipy.stats with overloading of addition and multiplication.

    Args:
            dtype (str): Data type of the distribution as a string. It must be one defined in `scipy.stats`
            loc (float): The location of the distribution (corresponds to mean in Gaussian)
            scale (float): The _scale of the distribution (corresponds to standard deviation in Gaussian)
            multipliers: An iterable of other distributon to *multiply* with this one
            adders: An iterable of other utility_priors to *add* to this one
            **kwargs:

    Examples:

        >>> d2 = Distribution('uniform')
        >>> print(d2.mean())
        0.5

        >>> try:
        ...     d = Distribution('something crazy')
        ... except ValueError as e:
        ...     print(str(e))
        Unknown distribution something crazy

    """

    def __init__(
        self,
        dtype: str,
        **kwargs,
    ) -> None:
        super().__init__()
        dist = getattr(stats, dtype.lower(), None)
        if dist is None:
            raise ValueError(f'Unknown distribution {dtype}')
        if 'loc' not in kwargs.keys():
            kwargs['loc'] = 0.0
        if 'scale' not in kwargs.keys():
            kwargs['scale'] = 1.0

        self.dist = dist(**kwargs)
        self.dtype = dtype
        self.__cached = None

    @classmethod
    def around(cls, value: float = 0.5, range: Tuple[float, float] = (0.0, 1.0),
               uncertainty: float = 0.5) -> 'Distribution':
        """
        Generates a uniform distribution around the input value in the given range with given uncertainty

        Args:
            value: The value to generate the distribution around
            range: The range of possible values
            uncertainty: The uncertainty level required. 0.0 means no uncertainty and 1.0 means full uncertainty

        Returns:
            Distribution A uniform distribution around `value` with uncertainty (scale) `uncertainty`
        """
        if uncertainty >= 1.0:
            return cls(dtype='uniform', loc=range[0], scale=range[1])
        if uncertainty <= 0.0:
            return cls(dtype='uniform', loc=value, scale=0.0)
        scale = uncertainty * (range[1] - range[0])
        loc = max(range[0], (random.random() - 1.0) * scale + value)
        if loc + scale > range[1]:
            loc -= (loc + scale - range[1])
        return cls(dtype='uniform', loc=loc, scale=scale)

    def mean(self) -> float:
        if self.dtype != 'uniform':
            raise NotImplementedError('Only uniform distributions are supported for now')
        if self.scale < 1e-6:
            return self.loc
        mymean = self.dist.mean()
        return float(mymean)

    def __float__(self):
        return float(self.mean())

    def __and__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return float(other)
        if self.dtype == 'uniform':
            beg = max(self.loc, other.loc)
            end = min(self.scale + self.loc, other.loc + other.scale)
            return Distribution(self.dtype
                                , loc=beg
                                , scale=end - beg
                                )
        raise NotImplementedError()

    def __or__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return float(other)
        if self.dtype == 'uniform':
            raise NotImplementedError('Current implementation assumes an overlap otherwise a mixture must be returned')
            beg = min(self.loc, other.loc)
            end = max(self.scale + self.loc, other.loc + other.scale)
            return Distribution(self.dtype
                                , loc=beg
                                , scale=end - beg
                                )
        raise NotImplementedError()

    def prob(self, val: float) -> float:
        """Returns the probability for the given value
        """
        return self.dist.prob(val)

    def sample(self, size: int = 1) -> np.ndarray:
        return self.dist.rvs(size=size)

    @property
    def loc(self):
        return self.dist.kwds.get('loc', 0.0)

    @property
    def scale(self):
        return self.dist.kwds.get('scale', 0.0)

    def __str__(self):
        if self.dtype == 'uniform':
            return f'U({self.loc}, {self.loc+self.scale})'
        return f'{self.dtype}(loc:{self.loc}, scale:{self.scale})'

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    __repr__ = __str__

    def __sub__(self, other):
        return float(self) - other

    def __add__(self, other):
        return float(self) + other

    def __radd__(self, other):
        return float(self) + other

    def __mul__(self, other):
        return float(self) * float(other)

    def __rmul__(self, other):
        return float(other) * float(self)

    def __divmod__(self, other):
        return float(self).__divmod__(other)


_inflect_engine = inflect.engine()


class ConfigReader:

    @classmethod
    def _parse_children_config(cls, children, scope):
        """Parses children in the given scope"""
        remaining_children = {}
        myconfig = {}
        setters = []
        for key, v in children.items():
            k, class_name = cls._split_key(key)
            if isinstance(v, Dict):
                if class_name is None:
                    class_name = stringcase.pascalcase(k)
                the_class = get_class(class_name=class_name, scope=scope)
                obj, obj_children = the_class.from_config(config=v, ignore_children=False, try_parsing_children=True
                                                                        , scope=scope)
                if obj_children is not None and len(obj_children) > 0:
                    remaining_children[k] = obj_children
                setter_name = 'set_' + k
                if hasattr(cls, setter_name):
                    setters.append((setter_name, obj))
                else:
                    myconfig[k] = obj
            elif isinstance(v, Iterable) and not isinstance(v, str):
                singular = _inflect_engine.singular_noun(k)
                if singular is False:
                    singular = k
                if class_name is None:
                    class_name = stringcase.pascalcase(singular)
                setter_name = 'set_' + k
                objs = []
                for current in list(v):
                    the_class = get_class(class_name=class_name, scope=scope)
                    obj = the_class.from_config(config=current, ignore_children=True
                                                              , try_parsing_children=True, scope=scope)
                    objs.append(obj)
                if hasattr(cls, setter_name):
                    setters.append((setter_name, objs))
                else:
                    myconfig[k] = objs
            else:
                # not a dictionary and not an iterable.
                remaining_children[k] = v

        return myconfig, remaining_children, setters

    @classmethod
    def _split_key(cls, key: str) -> Tuple[str, Optional[str]]:
        """Splits the key into a key name and a class name

        Remarks:

            - Note that if the given key has multiple colons the first two will be parsed as key name: class name and
              the rest will be ignored. This can be used to add comments

        """
        keys = key.split(':')
        if len(keys) == 1:
            return keys[0], None
        else:
            return keys[0], keys[1]

    @classmethod
    def read_config(cls, config: Union[str, dict], section: str = None) -> Dict[str, Any]:
        """
        Reads the configuration from a file or a dict and prepares it for parsing

        Args:
            config: Either a file name or a dictionary
            section: A section in the file or a key in the dictionary to use for loading params

        Returns:

            A dict ready to be parsed by from_config

        Remarks:


        """
        if isinstance(config, str):
            # If config is a string, assume it is a file and read it from the appropriate location
            def exists(nm):
                return os.path.exists(nm) and not os.path.isdir(nm)

            if not exists(config):
                name = pathlib.Path('./') / pathlib.Path(config)
                if exists(name):
                    config = str(name.absolute())
                else:
                    name = (pathlib.Path('./.negmas') / config).absolute()
                    if exists(name):
                        config = str(name)
                    else:
                        name = (pathlib.Path(os.path.expanduser('~/.negmas')) / config).absolute()
                        if exists(name):
                            config = str(name)
                        else:
                            raise ValueError(f'Cannot find config in {config}.')
            with open(config, 'r') as f:
                if config.endswith('.json'):
                    config = json.load(f)
                elif config.endswith('.cfg'):
                    config = eval(f.read())
                elif config.endswith('.yaml') or config.endswith('.yml'):
                    config = yaml.safe_load(f)
                else:
                    raise ValueError(f'Cannot parse {config}')

        if section is not None:
            config = config[section]  # type: ignore

        return config # type: ignore

    @classmethod
    def from_config(cls
                    , config: Union[str, dict]
                    , section: str = None
                    , ignore_children: bool = True
                    , try_parsing_children: bool = True
                    , scope=None):
        """
        Creates an object of this class given the configuration info

        Args:
            config: Either a file name or a dictionary
            section: A section in the file or a key in the dictionary to use for loading params
            ignore_children: If true then children will be ignored and there will be a single return
            try_parsing_children: If true the children will first be parsed as `ConfigReader` classes if they are not
            simple types (e.g. int, str, float, Iterable[int|str|float]
            scope: The scope at which to evaluate any child classes. This MUST be passed as scope=globals() if you are
            having any children that are to be parsed.

        Returns:

            An object of cls if ignore_children is True or a tuple with an object of cls and a dictionary with children
            that were not parsed.

        Remarks:

            - This function will return an object of its class after passing the key-value pairs found in the config to
              the init function.

            - Requiring passing scope=globals() to this function is to get around the fact that in python eval() will be
              called with a globals dictionary based on the module in which the function is defined not called. This means
              that in general when eval() is called to create the children, it will not have access to the class
              definitions of these children (except if they happen to be imported in this file). To avoid this problem
              causing an undefined_name exception, the caller must pass her globals() as the scope.

        """
        config = cls.read_config(config=config, section=section)

        if config is None:
            if ignore_children:
                return None
            else:
                return None, {}

        # now we have a dict called config which has our configuration

        myconfig = {}   # parts of the config that can directly be parsed
        children = {}   # parts of the config that need further parsing
        setters = []  # the setters are those configs that have a set_ function for them.

        def _is_simple(x):
            """Tests whether the input can directly be parsed"""
            return x is None or isinstance(x, int) or isinstance(x, str) or isinstance(x, float) or \
                   (isinstance(x, Iterable) and not isinstance(x, dict) and all(_is_simple(_) for _ in list(x)))

        def _set_simple_config(key, v) -> Optional[Dict[str, Any]]:
            """Sets a simple value v for key taken into accout its class and the class we are constructing"""
            key_name, class_name = cls._split_key(key)
            _setter = 'set_' + key_name
            params = {}
            if hasattr(cls, _setter):
                setters.append((_setter, v))
                return None
            params[key_name] = v if class_name is None else get_class(class_name=class_name, scope=scope)(v)
            return params

        # read the configs key by key and try to parse anything that is simple enough to parse

        for k, v in config.items(): # type: ignore
            if isinstance(v, Dict):
                children[k] = v
            elif isinstance(v, Iterable) and not isinstance(v, str):
                lst = list(v)
                if all(_is_simple(_) for _ in lst):
                    # that is a simple value of the form k:class_name = v. We construct class_name (if it exists) with v
                    # notice that we need to remove class_name when setting the key in myconfig
                    val = _set_simple_config(k, v)
                    if val is not None:
                        myconfig.update(val)
                else:
                    children[k] = v # type: ignore
            else:
                # that is a simple value of the form k:class_name = v. We construct class_name (if it exists) with v
                val = _set_simple_config(k, v)
                if val is not None:
                    myconfig.update(val)

        # now myconfig has all simply parsed parts and children has all non-parsed parts

        if len(children) > 0 and try_parsing_children:
            if scope is None:
                ValueError(f'scope is None but that is not allowed. You must pass scope=globals() or scope=locals() to '
                           f'from_config. If your classes are defined in the global scope pass globals() and if they '
                           f'are defined in local scope then pass locals(). You can only pass scope=None if you are '
                           f'sure that all of the constructor parameters of the class you are creating are simple '
                           f'values like ints floats and strings.')
            parsed_conf, remaining_children, setters = cls._parse_children_config(children=children
                                                                                  , scope=scope)
            myconfig.update(parsed_conf)
            children = remaining_children

        main_object = cls(**myconfig) # type: ignore

        if try_parsing_children:
            # we will only have setters if we have children
            for setter, value in setters:
                getattr(main_object, setter)(value)

        if ignore_children:
            return main_object
        return main_object, children


class Proxy:
    """A general proxy class."""

    def __init__(self, obj):
        self._obj = obj

    def __getattr__(self, item):
        return getattr(self._obj, item)


def get_full_type_name(t: Union[Type[Any], Callable]) -> str:
    """Gets the ful typename of a type. You *should not* pass an instance to this function but it may just work."""
    if not hasattr(t, '__module__') and not hasattr(t, '__name__'):
        t = type(t)
    return t.__module__ + '.' + t.__name__


def import_by_name(full_name: str) -> Any:
    """Imports something form a module using its full name"""
    if not isinstance(full_name, str):
        return full_name
    modules: List[str] = []
    parts = full_name.split('.')
    modules = parts[:-1]
    module_name = '.'.join(modules)
    item_name = parts[:-1]
    if len(modules) < 1:
        raise ValueError(f'Cannot get the object {item_name} in module {module_name}  (modules {modules})')
    module = importlib.import_module(module_name)
    return getattr(module, item_name)


def get_class(class_name: Union[str, Type], module_name: str = None, scope: dict = None) -> Type:
    """Imports and creates a class object for the given class name"""
    if not isinstance(class_name, str):
        return class_name
    modules: List[str] = []
    if module_name is not None:
        modules = module_name.split('.')
    modules += class_name.split('.')
    if len(modules) < 1:
        raise ValueError(f'Cannot get the class {class_name} in module {module_name}  (modules {modules})')
    class_name = stringcase.pascalcase(modules[-1])
    if len(modules) < 2:
        return eval(class_name, scope)
    module_name = '.'.join(modules[:-1])
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def instantiate(class_name: Union[str, Type], module_name: str = None, scope: dict = None, **kwargs) -> Any:
    """Imports and instantiates an object of a class"""
    return get_class(class_name, module_name)(**kwargs)


def humanize_time(secs, align=False, always_show_all_units=False):
    """
    Prints time that is given as seconds in human readable form. Useful only for times >=1sec.

    :param secs: float: number of seconds
    :param align: bool, optional: whether to align outputs so that they all take the same size (not implemented)
    :param always_show_all_units: bool, optional: Whether to always show days, hours, and minutes even when they
                                are zeros. default False
    :return: str: formated string with the humanized form
    """
    units = [("d", 86400), ("h", 3600), ("m", 60), ("s", 1)]
    parts = []
    for unit, mul in units:
        if secs / mul >= 1 or mul == 1 or always_show_all_units:
            if mul > 1:
                n = int(math.floor(secs / mul))
                secs -= n * mul
            else:
                n = secs if secs != int(secs) else int(secs)
            if align:
                parts.append("%2d%s%s" % (n, unit, ""))
            else:
                parts.append("%2d%s%s" % (n, unit, ""))
    return ":".join(parts)


def dump(d: Any, file_name: Union[str, os.PathLike]) -> None:
    """
    Saved an object depending on the extension of the file given. If the filename given has no extension,
    `DEFAULT_DUMP_EXTENSION` will be used

    Args:
        d: Object to save
        file_name: file name

    Remarks:

        - Supported formats are json, yaml
        - If None is given, the file will be created but will be empty

    """
    file_name = pathlib.Path(file_name).expanduser().absolute()
    if file_name.suffix == '':
        file_name = pathlib.Path(str(file_name) + '.' + DEFAULT_DUMP_EXTENSION)

    if d is None:
        with open(file_name, 'w') as f:
            pass

    if file_name.suffix == '.json':
        with open(file_name, 'w') as f:
            json.dump(d, f, sort_keys=True, indent=2)
    elif file_name.suffix == '.yaml':
        with open(file_name, 'w') as f:
            yaml.safe_dump(d, f)
    else:
        raise ValueError(f'Unkown extension {file_name.suffix} for {file_name}')
